from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import logging
from typing import Any, Dict, List, Optional, Union, cast

import mitzu.adapters.generic_adapter as GA
import mitzu.model as M
import pandas as pd
import sqlparse
import mitzu.helper as H

import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP
import sqlalchemy.sql.sqltypes as SA_T
from sqlalchemy.orm import aliased
from sqlalchemy.sql.type_api import TypeEngine


def fix_col_index(index: int, col_name: str):
    return col_name + f"_{index}"


COLUMN_NAME_REPLACE_STR = "___"


FieldReference = Union[SA.Column, EXP.Label]
SAMPLED_SOURCE_CTE_NAME = "sampled_source"


SIMPLE_TYPE_MAPPINGS = {
    SA_T.Numeric: M.DataType.NUMBER,
    SA_T.Integer: M.DataType.NUMBER,
    SA_T.Boolean: M.DataType.BOOL,
    SA_T.DateTime: M.DataType.DATETIME,
    SA_T.Date: M.DataType.DATETIME,
    SA_T.Time: M.DataType.DATETIME,
    SA_T.String: M.DataType.STRING,
}


def format_query(raw_query: Any):
    if type(raw_query) != str:
        raw_query = str(raw_query.compile(compile_kwargs={"literal_binds": True}))

    return sqlparse.format(raw_query, reindent=True, keyword_case="upper")


@dataclass
class SegmentSubQuery:
    event_data_table: M.EventDataTable
    table_ref: SA.Table
    where_clause: Any
    event_name: Optional[str] = None
    unioned_with: Optional[SegmentSubQuery] = None

    def union_all(self, sub_query: Optional[SegmentSubQuery]) -> SegmentSubQuery:
        if sub_query is None:
            return self
        uwith = self.unioned_with
        curr = self
        while uwith is not None:
            curr = uwith
            uwith = curr.unioned_with

        curr.unioned_with = sub_query
        return self


class SQLAlchemyAdapterError(Exception):
    pass


@dataclass
class SQLAlchemyAdapter(GA.GenericDatasetAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)
        self._table_cache: Dict[str, SA.Table] = {}
        self._connection: SA.engine.Connection = None
        self._engine: SA.engine.Engine = None

    def get_event_name_field(
        self,
        ed_table: M.EventDataTable,
        sa_table: Union[SA.table, EXP.CTE] = None,
    ) -> Any:
        if ed_table.event_name_field is not None:
            return self.get_field_reference(
                ed_table.event_name_field, ed_table, sa_table
            )
        else:
            return SA.literal(ed_table.event_name_alias)

    def get_conversion_sql(self, metric: M.ConversionMetric) -> str:
        return format_query(self._get_conversion_select(metric))

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        return self.execute_query(self._get_conversion_select(metric))

    def get_segmentation_sql(self, metric: M.SegmentationMetric) -> str:
        return format_query(self._get_segmentation_select(metric))

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        return self.execute_query(self._get_segmentation_select(metric))

    def get_retention_sql(self, metric: M.RetentionMetric) -> str:
        return format_query(self._get_retention_select(metric))

    def get_retention_df(self, metric: M.RetentionMetric) -> pd.DataFrame:
        return self.execute_query(self._get_retention_select(metric))

    def map_type(self, sa_type: Any) -> M.DataType:
        for sa_t, data_type in SIMPLE_TYPE_MAPPINGS.items():
            if issubclass(type(sa_type), sa_t):
                return data_type

        raise ValueError(f"{sa_type}[{type(sa_type)}]: is not supported.")

    def keep_alive_connection(self) -> bool:
        return False

    def execute_query(self, query: Any) -> pd.DataFrame:
        engine = self.get_engine()
        try:
            if H.LOGGER.isEnabledFor(logging.DEBUG):
                H.LOGGER.debug(f"Query:\n{format_query(query)}")
            if self._connection is None:
                self._connection = engine.connect()
            cursor_result = self._connection.execute(query)
            columns = cursor_result.keys()
            fetched = cursor_result.fetchall()
            if len(fetched) > 0:
                pdf = pd.DataFrame(fetched)
                pdf.columns = columns
            else:
                pdf = pd.DataFrame(columns=columns)
            return pdf
        except Exception as exc:
            self._connection = None
            H.LOGGER.error(f"Failed Query:\n{format_query(query)}")
            raise exc
        finally:
            if not self.keep_alive_connection():
                self._connection = None

    def get_engine(self) -> SA.engine.Engine:
        con = self.project.connection
        if self._engine is None:
            if con.url is None:
                url = self._get_connection_url(con)
            else:
                url = con.url
            self._engine = SA.create_engine(url)
        return self._engine

    def get_table_by_name(self, schema: str, table_name: str) -> SA.Table:
        full_name = f"{schema}.{table_name}"
        try:
            engine = self.get_engine()
        except Exception as e:
            raise SQLAlchemyAdapterError(f"Failed to connect to {full_name}") from e

        if full_name not in self._table_cache:
            metadata_obj = SA.MetaData()
            self._table_cache[full_name] = SA.Table(
                table_name,
                metadata_obj,
                schema=schema,
                autoload_with=engine,
                autoload=True,
            )
        return self._table_cache[full_name]

    def get_table(self, event_data_table: M.EventDataTable) -> SA.Table:
        if event_data_table.schema is not None:
            schema = event_data_table.schema
        elif self.project.connection.schema is not None:
            schema = self.project.connection.schema
        else:
            raise ValueError("Event data table doesn't have schema defined")
        return self.get_table_by_name(schema, event_data_table.table_name)

    def get_field_reference(
        self,
        field: M.Field,
        event_data_table: M.EventDataTable = None,
        sa_table: Union[SA.Table, EXP.CTE] = None,
    ) -> FieldReference:
        if sa_table is None and event_data_table is not None:
            sa_table = self.get_table(event_data_table)
        if sa_table is None:
            raise ValueError("Either sa_table or event_data_table has to be provided")

        if field._parent is None:
            return sa_table.columns.get(field._name)
        if field._parent._type == M.DataType.MAP:
            return SA.literal_column(
                f"{sa_table.name}.{field._parent._get_name()}['{field._name}']"
            )
        else:
            return SA.literal_column(f"{sa_table.name}.{field._get_name()}")

    def _get_struct_type(self) -> TypeEngine:
        raise NotImplementedError(
            "Generic SQL Alchemy Adapter doesn't support complex types (struct, row)"
        )

    def _parse_struct_type(self, sa_type: Any, name: str, path: str) -> M.Field:
        real_struct_type = self._get_struct_type()
        if isinstance(sa_type, real_struct_type):
            row: TypeEngine = sa_type
            sub_fields: List[M.Field] = []
            for n, st in row.attr_types:
                next_path = f"{path}.{n}"
                sf = self._parse_struct_type(
                    sa_type=st,
                    name=n,
                    path=next_path,
                )
                if sf._type == M.DataType and (
                    sf._sub_fields is None or len(sf._sub_fields) == 0
                ):
                    continue
                sub_fields.append(sf)
            return M.Field(
                _name=name, _type=M.DataType.STRUCT, _sub_fields=tuple(sub_fields)
            )
        else:
            return M.Field(_name=name, _type=self.map_type(sa_type))

    def _parse_map_type(
        self,
        sa_type: Any,
        name: str,
        event_data_table: M.EventDataTable,
    ) -> M.Field:
        raise NotImplementedError(
            "Generic SQL Alchemy Adapter doesn't support map types"
        )

    def _generate_retention_series_cte(
        self, start_dt: datetime, end_dt: datetime, time_window: M.TimeWindow
    ) -> EXP.CTE:
        selects = []
        curr_dt = start_dt
        index = 0
        while curr_dt <= end_dt:
            index_sel = SA.select(columns=[SA.literal(index).label(GA.RETENTION_INDEX)])
            selects.append(index_sel)
            index += time_window.value
            curr_dt = curr_dt + time_window.to_relative_delta()
        if len(selects) > 64:  # TODO: make this come from configs
            raise ValueError(
                "Too many retention periods to caluclate try reducing the scope"
            )
        return SA.union(*selects).cte()

    def list_schemas(self) -> List[str]:
        engine = self.get_engine()
        insp = SA.inspect(engine)
        return insp.get_schema_names()

    def list_tables(self, schema: str) -> List[str]:
        engine = self.get_engine()
        insp = SA.inspect(engine)
        return insp.get_table_names(schema=schema)

    def _flatten_fields(self, fields: List[M.Field]) -> List[M.Field]:
        """Flattens the tree field structure to a list.
            All leaf-nodes in the tree will be in the result list

        Args:
            fields (List[M.Field]): Fields that might contain subfields

        Returns:
            List[M.Field]: All subfields (with correct parents)
        """
        res = []
        for f in fields:
            if f._type.is_complex():
                if f._sub_fields is not None:
                    res.extend(self._flatten_fields(list(f._sub_fields)))
            else:
                res.append(f)
        return res

    def list_fields(self, event_data_table: M.EventDataTable) -> List[M.Field]:
        table = self.get_table(event_data_table)
        field_types = table.columns
        res = []
        for field_name, field_type in field_types.items():
            if field_name in event_data_table.ignored_fields:
                continue
            data_type = self.map_type(field_type.type)
            if data_type == M.DataType.STRUCT:
                complex_field = self._parse_struct_type(
                    sa_type=field_type.type,
                    name=field_name,
                    path=field_name,
                )
                if (
                    complex_field._sub_fields is None
                    or len(complex_field._sub_fields) == 0
                ):
                    continue
                res.append(complex_field)
            if data_type == M.DataType.MAP:
                map_field = self._parse_map_type(
                    sa_type=field_type.type,
                    name=field_name,
                    event_data_table=event_data_table,
                )
                if map_field._sub_fields is None or len(map_field._sub_fields) == 0:
                    continue
                res.append(map_field)
            else:
                field = M.Field(_name=field_name, _type=data_type)
                res.append(field)
        return self._flatten_fields(res)

    def list_all_table_columns(self, schema: str, table_name: str) -> List[M.Field]:
        table = self.get_table_by_name(schema, table_name)
        field_types = table.columns
        res = []
        for field_name, field_type in field_types.items():
            data_type = self.map_type(field_type.type)
            if data_type == M.DataType.STRUCT:
                complex_field = self._parse_struct_type(
                    sa_type=field_type.type,
                    name=field_name,
                    path=field_name,
                )
                if (
                    complex_field._sub_fields is None
                    or len(complex_field._sub_fields) == 0
                ):
                    continue
                res.append(complex_field)
            else:
                field = M.Field(_name=field_name, _type=data_type)
                res.append(field)
        return res

    def get_distinct_event_names(self, event_data_table: M.EventDataTable) -> List[str]:
        cte = aliased(
            self._get_dataset_discovery_cte(event_data_table),
            alias=SAMPLED_SOURCE_CTE_NAME,
            name=SAMPLED_SOURCE_CTE_NAME,
        )
        event_name_field = self.get_event_name_field(event_data_table, cte).label(
            GA.EVENT_NAME_ALIAS_COL
        )

        result = self.execute_query(
            SA.select(
                columns=[SA.distinct(event_name_field)],
            )
        )

        return pd.DataFrame(result)[GA.EVENT_NAME_ALIAS_COL].tolist()

    def _get_timewindow(self, timegroup: M.TimeGroup) -> Any:
        return SA.text(f"interval '1 {timegroup}'")

    def _get_datetime_interval(
        self, field_ref: FieldReference, timewindow: M.TimeWindow
    ) -> Any:
        return field_ref + SA.text(f"interval '{timewindow.value}' {timewindow.period}")

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        return SA.func.date_add(
            SA.text(f"'{time_group.name.lower()}'"), value_field_ref, field_ref
        )

    def _get_connection_url(self, con: M.Connection):
        user_name = "" if con.user_name is None else con.user_name
        password = "" if con.password is None else f":{con.password}"
        host_str = "" if con.host is None else str(con.host)
        if con.user_name is not None and con.host is not None:
            host_str = f"@{host_str}"
        port_str = "" if con.port is None else ":" + str(con.port)
        schema_str = "" if con.schema is None else f"/{con.schema}"
        url_params_str = "" if con.url_params is None else con.url_params
        if url_params_str != "" and url_params_str[0] != "?":
            url_params_str = "?" + url_params_str
        catalog_str = "" if con.catalog is None else f"/{con.catalog}"

        protocol = con.connection_type.value.lower()
        res = f"{protocol}://{user_name}{password}{host_str}{port_str}{catalog_str}{schema_str}{url_params_str}"
        return res

    def _get_distinct_array_agg_func(self, field_ref: FieldReference) -> Any:
        return SA.func.cast(SA.func.array_agg(SA.distinct(field_ref)), SA.JSON)

    def _column_index_support(self):
        return True

    def _get_dataset_discovery_cte(self, event_data_table: M.EventDataTable) -> EXP.CTE:
        if event_data_table.discovery_settings is None:
            raise ValueError("Missing discovery settings")

        table = self.get_table(event_data_table).alias("_evt")
        event_name_field = self.get_event_name_field(event_data_table, table)
        dt_field = self.get_field_reference(
            field=event_data_table.event_time_field,
            event_data_table=event_data_table,
            sa_table=table,
        )
        date_partition_filter = self._get_date_partition_filter(
            event_data_table,
            table,
            self.project.get_default_discovery_start_dt(),
            self.project.get_default_end_dt(),
        )

        raw_cte: EXP.CTE = SA.select(
            columns=[
                *table.columns.values(),
                (SA.cast(SA.func.random() * 1367, SA.Integer) % 100).label("__sample"),
                SA.func.row_number()
                .over(partition_by=event_name_field, order_by=SA.func.random())
                .label("rn"),
            ],
            whereclause=(
                (
                    dt_field
                    >= self._correct_timestamp(
                        self.project.get_default_discovery_start_dt()
                    )
                )
                & (
                    dt_field
                    <= self._correct_timestamp(self.project.get_default_end_dt())
                )
                & date_partition_filter
            ),
        ).cte()

        sampling_condition = SA.literal(True)
        if event_data_table.discovery_settings.lookback_days > 0:
            sampling_condition = (
                raw_cte.columns["__sample"]
                < event_data_table.discovery_settings.property_sample_rate
            ) | (
                raw_cte.columns["rn"]
                <= self.project.discovery_settings.min_property_sample_size
            )
        return SA.select(
            columns=raw_cte.columns.values(), whereclause=sampling_condition
        ).cte()

    def _get_column_values_df(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> pd.DataFrame:
        if event_data_table.discovery_settings is None:
            raise ValueError("Missing discovery settings")

        event_specific_str = "event specific" if event_specific else "generic"
        H.LOGGER.debug(f"Discovering {event_specific_str} field enums")

        cte = aliased(
            self._get_dataset_discovery_cte(event_data_table),
            alias=SAMPLED_SOURCE_CTE_NAME,
            name=SAMPLED_SOURCE_CTE_NAME,
        )
        event_name_select_field = (
            self.get_event_name_field(event_data_table, cte).label(
                GA.EVENT_NAME_ALIAS_COL
            )
            if event_specific
            else SA.literal(M.ANY_EVENT_NAME).label(GA.EVENT_NAME_ALIAS_COL)
        )

        query = SA.select(
            group_by=(
                SA.literal(1)
                if self._column_index_support()
                else SA.text(GA.EVENT_NAME_ALIAS_COL)
            ),
            columns=[event_name_select_field]
            + [
                SA.case(
                    (
                        SA.func.count(
                            SA.distinct(self.get_field_reference(f, sa_table=cte))
                        )
                        < event_data_table.discovery_settings.max_enum_cardinality,
                        self._get_distinct_array_agg_func(
                            self.get_field_reference(f, sa_table=cte)
                        ),
                    ),
                    else_=SA.literal(None),
                ).label(f._get_name().replace(".", COLUMN_NAME_REPLACE_STR))
                for f in fields
            ],
        )
        df = self.execute_query(query)
        df = df.rename(
            # This is required for complext types, as aliasing with `.`
            # doesn't work. The `.` comes from the get _get_name()
            # This issue might appear elsewhere as well
            columns={
                k: k.replace(COLUMN_NAME_REPLACE_STR, ".") for k in list(df.columns)
            }
        )
        return df.set_index(GA.EVENT_NAME_ALIAS_COL)

    def get_field_enums(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> Dict[str, M.EventDef]:
        enums = self._get_column_values_df(
            event_data_table, fields, event_specific
        ).to_dict("index")
        res: Dict[str, M.EventDef] = {}
        for evt, values in enums.items():
            field_defs: Dict[M.Field, M.EventFieldDef] = {}
            for f in fields:
                field_values = values[f._get_name()]
                if (
                    (field_values is None)
                    or (len(field_values) == 1 and field_values[0])
                    or (len(field_values) > 1)
                ):
                    vals = [v for v in field_values if v] if field_values else None
                    field_defs[f] = M.EventFieldDef(
                        _event_name=evt,
                        _field=f,
                        _event_data_table=event_data_table,
                        _enums=vals,
                    )

            res[evt] = M.EventDef(
                _event_name=evt,
                _fields=field_defs,
                _event_data_table=event_data_table,
            )
        return res

    def _get_date_trunc(
        self, time_group: M.TimeGroup, field_ref: FieldReference
    ) -> Any:
        return SA.func.date_trunc(time_group.name, field_ref)

    def _get_simple_segment_condition(
        self, table: SA.Table, segment: M.SimpleSegment
    ) -> Any:
        left = cast(M.EventFieldDef, segment._left)
        ref = self.get_field_reference(left._field, sa_table=table)
        op = segment._operator
        if op == M.Operator.IS_NULL:
            return ref.is_(None)
        if op == M.Operator.IS_NOT_NULL:
            return ref.is_not(None)

        if segment._right is None:
            return SA.literal(True)

        if op == M.Operator.EQ:
            return ref == segment._right
        if op == M.Operator.NEQ:
            return ref != segment._right
        if op == M.Operator.GT:
            return ref > segment._right
        if op == M.Operator.LT:
            return ref < segment._right
        if op == M.Operator.GT_EQ:
            return ref >= segment._right
        if op == M.Operator.LT_EQ:
            return ref <= segment._right
        if op == M.Operator.LIKE:
            return ref.like(segment._right)
        if op == M.Operator.NOT_LIKE:
            return SA.not_(ref.like(segment._right))
        if op == M.Operator.ANY_OF:
            if len(segment._right) == 0:
                return SA.literal(True)
            return ref.in_(segment._right)
        if op == M.Operator.NONE_OF:
            if len(segment._right) == 0:
                return SA.literal(True)
            return SA.not_(ref.in_(segment._right))
        raise ValueError(f"Operator {op} is not supported by SQLAlchemy Adapter.")

    def _get_segment_sub_query(
        self, segment: M.Segment, metric: M.Metric, step: int
    ) -> SegmentSubQuery:
        if isinstance(segment, M.SimpleSegment):
            s = cast(M.SimpleSegment, segment)
            left = s._left
            edt = left._event_data_table
            table = self.get_table(edt)
            evt_name_col = self.get_event_name_field(edt)
            event_time_filter = self._get_timewindow_where_clause(
                edt, table, metric, step
            )

            event_name_filter = (
                (evt_name_col == left._event_name)
                if left._event_name != M.ANY_EVENT_NAME
                else SA.literal(True)
            )
            if s._operator is None:
                return SegmentSubQuery(
                    event_name=left._event_name,
                    event_data_table=left._event_data_table,
                    table_ref=table,
                    where_clause=(event_name_filter & event_time_filter),
                )
            else:
                return SegmentSubQuery(
                    event_name=left._event_name,
                    event_data_table=left._event_data_table,
                    table_ref=table,
                    where_clause=(
                        event_name_filter
                        & event_time_filter
                        & self._get_simple_segment_condition(table, segment)
                    ),
                )
        elif isinstance(segment, M.ComplexSegment):
            c = cast(M.ComplexSegment, segment)
            l_query = self._get_segment_sub_query(c._left, metric, step)
            r_query = self._get_segment_sub_query(c._right, metric, step)
            if c._operator == M.BinaryOperator.AND:
                if l_query.event_data_table != r_query.event_data_table:
                    raise Exception(
                        "And (&) operator can only be between the same events (e.g. page_view & page_view)"
                    )
                return SegmentSubQuery(
                    event_name=None,
                    event_data_table=l_query.event_data_table,
                    table_ref=l_query.table_ref,
                    where_clause=l_query.where_clause & r_query.where_clause,
                )
            else:
                if l_query.event_data_table == r_query.event_data_table:
                    merged = SegmentSubQuery(
                        event_data_table=l_query.event_data_table,
                        table_ref=l_query.table_ref,
                        where_clause=l_query.where_clause | r_query.where_clause,
                    )
                    merged = merged.union_all(l_query.unioned_with)
                    return merged.union_all(r_query.unioned_with)
                else:
                    return l_query.union_all(r_query)
        else:
            raise ValueError(f"Segment of type {type(segment)} is not supported.")

    def _has_edt_event_field(
        self,
        group_field: M.EventFieldDef,
        ed_table: M.EventDataTable,
    ) -> bool:
        event_name = group_field._event_name
        field = group_field._field
        dd: Optional[M.DiscoveredProject] = self.project._discovered_project.get_value()
        if dd is None:
            raise Exception("No DiscoveredProject was provided to SQLAlchemy Adapter.")
        events = dd.definitions.get(ed_table)
        if events is not None:
            event_def_reference = events.get(event_name)
            if event_def_reference is not None:
                event_def = event_def_reference.get_value_if_exists()
                for edt_evt_field in event_def._fields:
                    if edt_evt_field == field:
                        return True
        return False

    def _find_group_by_field_ref(
        self,
        group_field: M.EventFieldDef,
        sa_table: SA.Table,
        ed_table: M.EventDataTable,
    ):
        field = group_field._field
        if field._parent is None:
            columns = sa_table.columns
            if field._name in columns:
                return self.get_field_reference(field, sa_table=sa_table)
        elif self._has_edt_event_field(group_field, ed_table):
            return self.get_field_reference(field, sa_table=sa_table)
        return SA.literal(None)

    def _get_segment_sub_query_cte(
        self,
        sub_query: SegmentSubQuery,
        group_field: Optional[M.EventFieldDef] = None,
        resolution: M.Resolution = M.Resolution.EVERY_EVENT,
    ) -> EXP.CTE:
        selects = []
        while True:
            ed_table = sub_query.event_data_table

            group_by_col = SA.literal(None)
            if group_field is not None:
                group_by_col = self._find_group_by_field_ref(
                    group_field, sub_query.table_ref, ed_table
                )

            datetime_col = self.get_field_reference(ed_table.event_time_field, ed_table)
            resolution_tg = resolution.get_time_group()
            if resolution_tg is not None:
                datetime_col = self._get_date_trunc(
                    field_ref=datetime_col,
                    time_group=resolution_tg,
                )

            select = SA.select(
                columns=[
                    self.get_field_reference(ed_table.user_id_field, ed_table).label(
                        GA.CTE_USER_ID_ALIAS_COL
                    ),
                    datetime_col.label(GA.CTE_DATETIME_COL),
                    group_by_col.label(GA.CTE_GROUP_COL),
                ],
                whereclause=(sub_query.where_clause),
            )
            if resolution != M.Resolution.EVERY_EVENT:
                select = select.distinct()

            selects.append(select)
            if sub_query.unioned_with is not None:
                sub_query = sub_query.unioned_with
            else:
                break

        return SA.union_all(*selects).cte()

    def _correct_timestamp(self, dt: datetime) -> Any:
        return dt

    def _get_date_partition_filter(
        self,
        edt: M.EventDataTable,
        table: SA.Table,
        start_dt: datetime,
        end_dt: datetime,
    ):
        if edt.date_partition_field is not None:
            dt_part = SA.func.date(
                self.get_field_reference(edt.date_partition_field, edt, table)
            )
            return (dt_part >= SA.func.date(start_dt.date())) & (
                dt_part <= SA.func.date(end_dt.date())
            )
        else:
            return SA.literal(True)

    def _get_timewindow_where_clause(
        self, edt: M.EventDataTable, table: SA.Table, metric: M.Metric, step: int
    ) -> Any:
        evt_time_col = self.get_field_reference(edt.event_time_field, edt)
        start_date = metric._start_dt
        end_date = metric._end_dt

        if step > 0:
            if isinstance(metric, M.ConversionMetric):
                end_date = end_date + metric._conv_window.to_relative_delta()
            elif isinstance(metric, M.RetentionMetric):
                end_date = end_date + metric._retention_window.to_relative_delta()

        if metric._resolution == M.Resolution.ONE_USER_EVENT_PER_DAY:
            start_date = start_date.replace(hour=0, minute=0, second=0)
        elif metric._resolution == M.Resolution.ONE_USER_EVENT_PER_HOUR:
            start_date = start_date.replace(minute=0, second=0)
        elif metric._resolution == M.Resolution.ONE_USER_EVENT_PER_MINUTE:
            start_date = start_date.replace(second=0)

        start_date = start_date.replace(microsecond=0)
        end_date = end_date.replace(microsecond=0)
        event_time_filter = (evt_time_col >= self._correct_timestamp(start_date)) & (
            evt_time_col <= self._correct_timestamp(end_date)
        )
        date_partition_filter = self._get_date_partition_filter(
            edt, table, start_date, end_date
        )
        return event_time_filter & date_partition_filter

    def _get_seg_aggregation(self, metric: M.Metric, cte: EXP.CTE) -> Any:
        at = metric._agg_type
        if at == M.AggType.COUNT_EVENTS:
            return SA.func.count(cte.columns.get(GA.CTE_USER_ID_ALIAS_COL))
        elif at == M.AggType.COUNT_UNIQUE_USERS:
            return SA.func.count(cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct())
        else:
            raise ValueError(
                f"Aggregation type {at.name} is not supported for segmentation"
            )

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        at = metric._agg_type
        if at == M.AggType.CONVERSION:
            return (
                SA.func.count(cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct())
                * 100.0
                / SA.func.count(
                    first_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
                )
            )
        elif at == M.AggType.PERCENTILE_TIME_TO_CONV:
            raise NotImplementedError(
                "Sql Alchemy adapter doesn't support percentile calculation"
            )
        else:
            raise ValueError(f"Aggregation type {at} is not supported for conversion")

    def _get_segmentation_select(self, metric: M.SegmentationMetric) -> Any:
        sub_query = self._get_segment_sub_query(metric._segment, metric, step=0)
        cte: EXP.CTE = aliased(
            self._get_segment_sub_query_cte(sub_query, metric._group_by)
        )

        evt_time_group = (
            self._get_date_trunc(
                field_ref=cte.columns.get(GA.CTE_DATETIME_COL),
                time_group=metric._time_group,
            )
            if metric._time_group != M.TimeGroup.TOTAL
            else SA.literal(None)
        )

        group_by = (
            cte.columns.get(GA.CTE_GROUP_COL)
            if metric._group_by is not None
            else SA.literal(None)
        )

        return SA.select(
            columns=[
                evt_time_group.label(GA.DATETIME_COL),
                group_by.label(GA.GROUP_COL),
                self._get_seg_aggregation(metric, cte).label(GA.AGG_VALUE_COL),
            ],
            group_by=(
                [SA.literal(1), SA.literal(2)]
                if self._column_index_support()
                else [SA.text(GA.DATETIME_COL), SA.text(GA.GROUP_COL)]
            ),
        )

    def _get_conversion_select(self, metric: M.ConversionMetric) -> Any:
        first_segment = metric._conversion._segments[0]
        first_cte = self._get_segment_sub_query_cte(
            self._get_segment_sub_query(first_segment, metric, step=0),
            metric._group_by,
            metric._resolution,
        )
        first_group_by = (
            first_cte.columns.get(GA.CTE_GROUP_COL)
            if metric._group_by is not None
            else SA.literal(None)
        )

        time_group = metric._time_group
        if time_group != M.TimeGroup.TOTAL:
            first_evt_time_group = self._get_date_trunc(
                field_ref=first_cte.columns.get(GA.CTE_DATETIME_COL),
                time_group=time_group,
            )
        else:
            first_evt_time_group = SA.literal(None)

        other_segments = metric._conversion._segments[1:]

        steps = [first_cte]
        other_selects = []
        joined_source = first_cte
        for i, seg in enumerate(other_segments):
            prev_table = steps[i]
            prev_cols = prev_table.columns
            curr_cte = self._get_segment_sub_query_cte(
                self._get_segment_sub_query(seg, metric, step=i + 1),
                resolution=metric._resolution,
            )
            curr_cols = curr_cte.columns
            curr_used_id_col = curr_cols.get(GA.CTE_USER_ID_ALIAS_COL)

            steps.append(curr_cte)
            other_selects.extend(
                [
                    SA.func.count(
                        curr_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
                    ).label(fix_col_index(i + 2, GA.USER_COUNT_COL)),
                    self._get_conv_aggregation(metric, curr_cte, first_cte).label(
                        fix_col_index(i + 2, GA.AGG_VALUE_COL)
                    ),
                ]
            )
            joined_source = joined_source.join(
                curr_cte,
                (
                    (prev_cols.get(GA.CTE_USER_ID_ALIAS_COL) == curr_used_id_col)
                    & (
                        curr_cols.get(GA.CTE_DATETIME_COL)
                        > prev_cols.get(GA.CTE_DATETIME_COL)
                    )
                    & (
                        curr_cols.get(GA.CTE_DATETIME_COL)
                        <= self._get_datetime_interval(
                            first_cte.columns.get(GA.CTE_DATETIME_COL),
                            metric._conv_window,
                        )
                    )
                ),
                isouter=True,
            )

        columns = [
            first_evt_time_group.label(GA.DATETIME_COL),
            first_group_by.label(GA.GROUP_COL),
            SA.func.count(
                first_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
            ).label(fix_col_index(1, GA.USER_COUNT_COL)),
            self._get_conv_aggregation(metric, first_cte, first_cte).label(
                fix_col_index(1, GA.AGG_VALUE_COL)
            ),
        ]
        columns.extend(other_selects)
        return SA.select(
            columns=columns,
            group_by=(
                [SA.literal(1), SA.literal(2)]
                if self._column_index_support()
                else [SA.text(GA.DATETIME_COL), SA.text(GA.GROUP_COL)]
            ),
        ).select_from(joined_source)

    def _get_retention_select(self, metric: M.RetentionMetric) -> Any:
        initial_cte = self._get_segment_sub_query_cte(
            self._get_segment_sub_query(metric._initial_segment, metric, step=0),
            metric._group_by,
            metric._resolution,
        )

        retention_index_cte = self._generate_retention_series_cte(
            metric._start_dt, metric._end_dt, metric._retention_window
        ).alias("ret_indeces")

        initial_group_by = (
            initial_cte.columns.get(GA.CTE_GROUP_COL)
            if metric._group_by is not None
            else SA.literal(None)
        )

        time_group = (
            self._get_date_trunc(
                metric._time_group, initial_cte.columns.get(GA.CTE_DATETIME_COL)
            )
            if metric._time_group != M.TimeGroup.TOTAL
            else SA.literal(None)
        )

        retaining_cte = self._get_segment_sub_query_cte(
            self._get_segment_sub_query(metric._retaining_segment, metric, step=1),
            resolution=metric._resolution,
        )

        retention_interval_func = self._get_dynamic_datetime_interval(
            field_ref=initial_cte.columns.get(GA.CTE_DATETIME_COL),
            value_field_ref=retention_index_cte.columns.get(GA.RETENTION_INDEX),
            time_group=metric._retention_window.period,
        )

        joined_source = initial_cte.join(retention_index_cte, True).join(
            retaining_cte,
            (
                (
                    initial_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL)
                    == retaining_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL)
                )
                & (
                    self._get_datetime_column(retaining_cte, GA.CTE_DATETIME_COL)
                    > retention_interval_func
                )
                & (
                    self._get_datetime_column(retaining_cte, GA.CTE_DATETIME_COL)
                    <= self._get_datetime_interval(
                        retention_interval_func,
                        metric._retention_window,
                    )
                )
            ),
            isouter=True,
        )

        columns = [
            time_group.label(GA.DATETIME_COL),
            initial_group_by.label(GA.GROUP_COL),
            retention_index_cte.columns.get(GA.RETENTION_INDEX),
            SA.func.count(
                initial_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
            ).label(fix_col_index(1, GA.USER_COUNT_COL)),
            SA.func.count(
                retaining_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
            ).label(fix_col_index(2, GA.USER_COUNT_COL)),
            (
                SA.func.count(
                    retaining_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
                )
                * 100.0
                / SA.func.count(
                    initial_cte.columns.get(GA.CTE_USER_ID_ALIAS_COL).distinct()
                )
            ).label(GA.AGG_VALUE_COL),
        ]

        return SA.select(
            columns=columns,
            group_by=(
                [SA.literal(1), SA.literal(2), SA.literal(3)]
                if self._column_index_support()
                else [
                    SA.text(GA.DATETIME_COL),
                    SA.text(GA.GROUP_COL),
                    SA.text(GA.RETENTION_INDEX),
                ]
            ),
        ).select_from(joined_source)

    def _get_datetime_column(self, cte: Any, name: str) -> Any:
        return cte.columns.get(name)

    def test_connection(self):
        self.execute_query(
            SA.select(columns=[SA.literal(True).label("test_connection")])
        )

    def stop_current_execution(self):
        if self._connection is not None:
            self._connection.connection.close()



from __future__ import annotations


import dash_bootstrap_components as dbc
from typing import Any, Dict, Optional

TOAST_INDEX_TYPE = "mitzu_toast"


def create_toast(
    id: str, duration: Optional[int] = 5000, position_style: Dict[str, Any] = None
) -> dbc.Toast:
    if position_style is None:
        position_style = {"top": 66, "right": 10}
    return dbc.Toast(
        children=[],
        id=id,
        is_open=False,
        header=[],
        dismissable=True,
        duration=duration,
        style={
            "position": "fixed",
            "z-index": "1000",
            "width": "fit-content",
            **position_style,
        },
    )



from __future__ import annotations

from typing import Any, Dict, List, Optional

import dash.development.base_component as bc
import mitzu.model as M
import mitzu.webapp.pages.explore.simple_segment_handler as SS
from dash import html
from mitzu.webapp.helper import CHILDREN, get_event_names, WITH_VALUE_CLS
from mitzu.helper import value_to_label
import dash_mantine_components as dmc

EVENT_SEGMENT = "event_segment"
EVENT_NAME_DROPDOWN = "event_name_dropdown"
SIMPLE_SEGMENT_CONTAINER = "simple_segment_container"


def create_event_name_dropdown(
    index,
    discovered_project: M.DiscoveredProject,
    step: int,
    event_segment_index: int,
    segment: Optional[M.Segment],
) -> dmc.Select:
    options = (
        [
            {"label": value_to_label(v), "value": v}
            for v in discovered_project.get_all_event_names()
        ]
        if discovered_project is not None
        else []
    )
    options.sort(key=lambda v: v["label"])

    if step == 0 and event_segment_index == 0:
        placeholder = "+ Select Event"
    elif step > 0 and event_segment_index == 0:
        placeholder = "+ Then"
    else:
        placeholder = "+ Select Another Event"

    segment_event_names = get_event_names(segment)
    if len(set(segment_event_names)) > 1:
        raise Exception("Multiple event complex segment for EventSegmentHandler")
    value = None if len(segment_event_names) == 0 else segment_event_names[0]

    return dmc.Select(
        data=options,
        value=value,
        searchable=True,
        clearable=True,
        className=(
            EVENT_NAME_DROPDOWN
            + " "
            + ("" if value is None else WITH_VALUE_CLS)
            + " border border-0 rounded-2"
        ),
        placeholder=placeholder,
        id={
            "type": EVENT_NAME_DROPDOWN,
            "index": index,
        },
    )


def get_event_simple_segments(segment: M.Segment) -> List[M.SimpleSegment]:
    if isinstance(segment, M.SimpleSegment):
        return [segment]
    elif isinstance(segment, M.ComplexSegment):
        if segment._operator != M.BinaryOperator.AND:
            raise Exception(
                f"Invalid event level complex segment operator: {segment._operator}"
            )
        return get_event_simple_segments(segment._left) + get_event_simple_segments(
            segment._right
        )
    else:
        raise Exception(f"Unsupported Segment Type: {type(segment)}")


def create_simple_segments_container(
    type_index: str,
    segment: Optional[M.Segment],
    discovered_project: M.DiscoveredProject,
) -> Optional[html.Div]:
    children = []
    if segment is None:
        return None
    event_simple_segments = get_event_simple_segments(segment)
    if isinstance(segment, M.ComplexSegment):
        for seg_index, ess in enumerate(event_simple_segments):
            children.append(
                SS.from_simple_segment(ess, discovered_project, type_index, seg_index)
            )
    elif isinstance(segment, M.SimpleSegment) and isinstance(
        segment._left, M.EventFieldDef
    ):
        children.append(
            SS.from_simple_segment(segment, discovered_project, type_index, 0)
        )

    event_name = event_simple_segments[0]._left._event_name
    event_def = discovered_project.get_event_def(event_name)
    if event_def is None:
        raise Exception(
            f"Invalid state, {event_name} is not possible to find in discovered datasource."
        )
    children.append(
        SS.from_simple_segment(
            M.SimpleSegment(event_def),
            discovered_project,
            type_index,
            len(children),
        )
    )

    return html.Div(
        id={"type": SIMPLE_SEGMENT_CONTAINER, "index": type_index}, children=children
    )


def from_segment(
    segment: Optional[M.Segment],
    discovered_project: M.DiscoveredProject,
    funnel_step: int,
    event_segment_index: int,
) -> bc.Component:
    type_index = f"{funnel_step}-{event_segment_index}"
    event_dd = create_event_name_dropdown(
        type_index, discovered_project, funnel_step, event_segment_index, segment
    )
    children = [event_dd]
    simples_segs_container = create_simple_segments_container(
        type_index, segment, discovered_project
    )
    if simples_segs_container is not None:
        children.append(simples_segs_container)

    component = html.Div(
        id={"type": EVENT_SEGMENT, "index": type_index},
        children=children,
        className=EVENT_SEGMENT,
    )
    return component


def from_all_inputs(
    discovered_project: Optional[M.DiscoveredProject],
    event_segment: Dict[str, Any],
) -> Optional[M.Segment]:
    if discovered_project is None:
        return None
    event_name_dd_value = event_segment.get(EVENT_NAME_DROPDOWN)
    simple_segments = event_segment.get(CHILDREN, {})

    if event_name_dd_value is None:
        return None

    res_segment: Optional[M.Segment] = None
    for _, simple_segment in simple_segments.items():
        simple_seg = SS.from_all_inputs(discovered_project, simple_segment)

        if simple_seg is None:
            continue
        if simple_seg._left._event_name != event_name_dd_value:
            continue

        if res_segment is None:
            res_segment = simple_seg
        else:
            res_segment = res_segment & simple_seg

    if res_segment is None and event_name_dd_value is not None:
        return M.SimpleSegment(
            _left=discovered_project.get_event_def(event_name_dd_value)
        )
    return res_segment



import mitzu.model as M
from dash import html

PROP_CONNECTION = "connection"
PROJECT_INDEX_TYPE = "project_property"
EDT_TBL_BODY = "edt_table_body"

MISSING_VALUE = "<Required>"
MISSING_FIELD = M.Field(MISSING_VALUE, M.DataType.STRING)


def create_empty_edt(schema: str, table_name: str):
    return M.EventDataTable(
        table_name=table_name,
        schema=schema,
        user_id_field=MISSING_FIELD,
        event_time_field=MISSING_FIELD,
    )


def get_value_from_row(tr: html.Tr, col: int) -> str:
    return tr.get("props").get("children")[col].get("props")["children"]



from typing import cast, List, Dict, Any

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import flask
from dash import html, register_page, callback, Output, Input, ALL, ctx, no_update, dcc

import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.paths as P
import mitzu.webapp.navbar as NB
from mitzu.webapp.auth.decorator import restricted_layout, restricted
from mitzu.webapp.helper import TBL_CLS, TBL_HEADER_CLS
import mitzu.webapp.cache as C
import pickle
import json
import traceback
import base64


DELETE_BUTTON_TYPE = "admin_delete_button_type"
STORAGE_TABLE = "admin_storage_table"
TBL_BODY_CONTAINER = "admin_table_body_container"
STORAGE_DOWNLOAD = "admin_storage_download"
STORAGE_RESET = "admin_reset"
ADMIN_NAVBAR = "admin_navbar"

STORAGE_DOWNLOAD_BUTTON = "admin_storage_download_button"
STORAGE_RESET_BUTTON = "admin_storage_reset_button"
CLEAR_LOCAL_CACHE_BUTTON = "clear_local_cache_button"
STORAGE_INFO = "admin_storage_info"


def create_table_body_rows(mitzu_cache: C.MitzuCache) -> List[html.Td]:
    all_keys = mitzu_cache.list_keys()
    all_keys = sorted(all_keys)
    rows = []
    for key in all_keys:
        rows.append(
            html.Tr(
                [
                    html.Td(key),
                    html.Td(
                        [
                            dbc.Button(
                                "Delete",
                                size="sm",
                                color="danger",
                                outline=True,
                                id={"type": DELETE_BUTTON_TYPE, "index": key},
                            ),
                        ],
                    ),
                ],
                className=TBL_CLS,
            )
        )
    return rows


@restricted_layout
def layout() -> bc.Component:
    cache = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)).cache

    table_header = html.Thead(
        html.Tr([html.Th("key"), html.Th("value")], className=TBL_HEADER_CLS)
    )
    rows = create_table_body_rows(cache)

    table = dbc.Table(
        [table_header, html.Tbody(rows, id=TBL_BODY_CONTAINER)],
        id=STORAGE_TABLE,
        bordered=True,
        hover=True,
        responsive=True,
        striped=True,
        size="sm",
    )
    return html.Div(
        [
            NB.create_mitzu_navbar(id=ADMIN_NAVBAR, create_explore_button=False),
            dcc.Download(id=STORAGE_DOWNLOAD),
            dbc.Container(
                children=[
                    html.H4("Mitzu Storage Content:", className="card-title mb-3"),
                    dbc.Row(
                        [
                            dbc.Col(
                                dbc.Button(
                                    [
                                        html.B(className="bi bi-cloud-download me-1"),
                                        "Download json",
                                    ],
                                    id=STORAGE_DOWNLOAD_BUTTON,
                                    color="primary",
                                    size="sm",
                                ),
                                lg="auto",
                                sm=12,
                                class_name="mb-1",
                            ),
                            dbc.Col(
                                dcc.Upload(
                                    dbc.Button(
                                        [
                                            html.B(className="bi bi-cloud-upload me-1"),
                                            "Reset from json",
                                        ],
                                        id=STORAGE_RESET_BUTTON,
                                        color="info",
                                        size="sm",
                                    ),
                                    accept="application/json",
                                    multiple=False,
                                    id=STORAGE_RESET,
                                ),
                                lg="auto",
                                sm=12,
                                class_name="mb-1",
                            ),
                            dbc.Col(
                                children=[],
                                id=STORAGE_INFO,
                            ),
                        ]
                    ),
                    html.Hr(),
                    table,
                ]
            ),
        ]
    )


def validate_input_json(storage: Dict[str, str]) -> Dict[str, Any]:
    res: Dict[str, Any] = {}
    for k, v in storage.items():
        unpickled = pickle.loads(base64.b64decode(v.encode("ascii")))
        if unpickled is None:
            raise ValueError(f"Invalid values for key: {k}")
        res[k] = unpickled
    return res


@callback(
    Output(TBL_BODY_CONTAINER, "children"),
    Output(STORAGE_INFO, "children"),
    Input({"type": DELETE_BUTTON_TYPE, "index": ALL}, "n_clicks"),
    Input(STORAGE_RESET, "contents"),
    prevent_initial_call=True,
    background=True,
    running=[
        [Output(STORAGE_DOWNLOAD_BUTTON, "disabled"), True, False],
        [Output(STORAGE_RESET_BUTTON, "disabled"), True, False],
    ],
    interval=100,
)
@restricted
def update_tbl_body(delete_n_clicks: int, content: str):
    try:
        if ctx.triggered_id == STORAGE_RESET and content is not None:
            cache = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            ).cache
            _, content_string = content.split(",")
            decoded = base64.b64decode(content_string)
            storage = json.loads(decoded.decode())
            to_restore = validate_input_json(storage)
            cache.clear_all()
            for k, v in to_restore.items():
                cache.put(k, v)
            res = create_table_body_rows(cache)
            return res, ""
        elif ctx.triggered_id != STORAGE_RESET and delete_n_clicks is not None:
            cache = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            ).cache
            key = ctx.triggered_id["index"]
            cache.clear(key)
            return create_table_body_rows(cache), ""
        else:
            return no_update, no_update
    except Exception as exc:
        traceback.print_exc()
        return no_update, html.P(f"{str(exc)}", className="text-danger lead")


@callback(
    Output(STORAGE_DOWNLOAD, "data"),
    Input(STORAGE_DOWNLOAD_BUTTON, "n_clicks"),
    prevent_initial_call=True,
    background=True,
    running=[
        [Output(STORAGE_DOWNLOAD_BUTTON, "disabled"), True, False],
        [Output(STORAGE_RESET_BUTTON, "disabled"), True, False],
    ],
)
@restricted
def download_button_clicked(n_clicks: int):
    if n_clicks is not None:
        cache = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        ).cache
        all_keys = cache.list_keys(prefix="__", strip_prefix=False)
        all_keys = sorted(all_keys)
        res: Dict[str, Any] = {}

        for k in all_keys:
            value = cache.get(k)
            res[k] = base64.b64encode(pickle.dumps(value)).decode("ascii")
        res_str = json.dumps(res)
        return dict(content=res_str, filename="mitzu_storage.json")
    else:
        return no_update


register_page(
    __name__,
    path=P.ADMIN_PAGE,
    title="Mitzu - Admin",
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple

import pyathena
import pyathena.connection as pa_conn
from mitzu.adapters.sqlalchemy.athena.sqlalchemy import compiler, datatype, error

from sqlalchemy.engine.base import Connection
from sqlalchemy.engine.default import DefaultDialect, DefaultExecutionContext
from sqlalchemy.engine.url import URL


class AthenaDialect(DefaultDialect):
    name = "athena"
    driver = "rest"

    statement_compiler = compiler.AthenaSQLCompiler
    ddl_compiler = compiler.AthenaDDLCompiler
    type_compiler = compiler.AthenaTypeCompiler
    preparer = compiler.AthenaIdentifierPreparer

    # Data Type
    supports_native_enum = False
    supports_native_boolean = True
    supports_native_decimal = True

    # Column options
    supports_sequences = False
    supports_comments = True
    inline_comments = True
    supports_default_values = False

    # DDL
    supports_alter = False

    # DML
    # Queries of the form `INSERT () VALUES ()` is not supported by Athena.
    supports_empty_insert = False
    supports_multivalues_insert = True
    postfetch_lastrowid = False

    # Caching
    # Warnings are generated by SQLAlchmey if this flag is not explicitly set
    # and tests are needed before being enabled
    supports_statement_cache = False

    @classmethod
    def dbapi(cls):
        return pyathena

    def create_connect_args(self, url: URL) -> Tuple[Sequence[Any], Mapping[str, Any]]:
        args: Sequence[Any] = list()
        kwargs: Dict[str, Any] = dict(server_hostname=url.host)

        if url.port:
            kwargs["port"] = url.port

        db_parts = (url.database or "system").split("/")

        if len(db_parts) == 1:
            kwargs["schema"] = db_parts[0]
        elif len(db_parts) == 2:
            kwargs["catalog"] = db_parts[0]
            kwargs["schema"] = db_parts[1]
        else:
            raise ValueError(f"Unexpected database format {url.database}")

        if url.password:
            kwargs["aws_access_key_id"] = url.username
            kwargs["aws_secret_access_key"] = url.password

        if "s3_staging_dir" in url.query:
            kwargs["s3_staging_dir"] = url.query["s3_staging_dir"]

        if "work_group" in url.query:
            kwargs["work_group"] = url.query["work_group"]

        kwargs["region_name"] = url.query["region"]
        kwargs["session"] = None

        return args, kwargs

    def get_columns(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[Dict[str, Any]]:

        db_conn: pa_conn.Connection = connection._dbapi_connection
        schema, table_name = self._handle_table_name(table_name, schema)
        metadata = db_conn.cursor().get_table_metadata(
            catalog_name=None, schema_name=schema, table_name=table_name
        )
        columns = []
        records = metadata.partition_keys + metadata.columns
        for record in records:
            if record.type is None:
                raise error.AthenaQueryError(f"'{record.name}' column type is unkown")
            column = dict(
                name=record.name,
                type=datatype.parse_sqltype(record.type),
                nullable=True,
                default=None,
            )
            columns.append(column)
        return columns

    def get_pk_constraint(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> Dict[str, Any]:
        """Athena has no support for primary keys. Returns a dummy"""
        return dict(name=None, constrained_columns=[])

    def get_primary_keys(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[str]:
        pk = self.get_pk_constraint(connection, table_name, schema)
        return pk.get("constrained_columns")  # type: ignore

    def get_foreign_keys(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[Dict[str, Any]]:
        """Athena has no support for foreign keys. Returns an empty list."""
        return []

    def get_schema_names(self, connection: Connection, **kw) -> List[str]:
        query = "show schemas"
        res = connection.execute(query)

        return [row[0] for row in res]

    def get_table_names(
        self, connection: Connection, schema: str = None, **kw
    ) -> List[str]:
        schema = schema or self._get_default_schema_name(connection)
        query = f"show tables in {schema}"
        res = connection.execute(query)
        return [row[0] for row in res]

    def get_temp_table_names(
        self, connection: Connection, schema: str = None, **kw
    ) -> List[str]:
        """Athena has no support for listing temporary tables. Returns an empty list."""
        return []

    def get_view_names(
        self, connection: Connection, schema: str = None, **kw
    ) -> List[str]:
        schema = schema or self._get_default_schema_name(connection)
        query = f"show views in {schema}"
        res = connection.execute(query)
        return [row.viewName for row in res]

    def get_temp_view_names(
        self, connection: Connection, schema: str = None, **kw
    ) -> List[str]:
        """Athena has no support for temporary views. Returns an empty list."""
        return []

    def get_view_definition(
        self, connection: Connection, view_name: str, schema: str = None, **kw
    ) -> str:
        schema, view_name = self._handle_table_name(view_name, schema)

        query = f"show create table {schema}.{view_name}"
        return connection.execute(query).scalar()

    def get_indexes(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[Dict[str, Any]]:
        """Athena has no support for indexes"""
        return []

    def get_sequence_names(
        self, connection: Connection, schema: str = None, **kw
    ) -> List[str]:
        """Athena has no support for sequences. Returns an empty list."""
        return []

    def get_unique_constraints(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[Dict[str, Any]]:
        """Athena has no support for unique constraints. Returns an empty list."""
        return []

    def get_check_constraints(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> List[Dict[str, Any]]:
        """Athena has no support for check constraints. Returns an empty list."""
        return []

    def get_table_comment(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> Dict[str, Any]:
        return {}

    def has_schema(self, connection: Connection, schema: str) -> bool:
        return schema in self.get_schema_names(connection)

    def has_table(
        self, connection: Connection, table_name: str, schema: str = None, **kw
    ) -> bool:
        schema, table_name = self._handle_table_name(table_name, schema)
        rows = connection.cursor().tables().fetchall()
        for row in rows:
            if (
                row.TABLE_NAME.lower() == table_name.lower()
                and row.TABLE_SCHEM.lower() == schema.lower()
            ):
                return True

        return False

    def has_sequence(
        self, connection: Connection, sequence_name: str, schema: str = None, **kw
    ) -> bool:
        """Athena has no support for sequence. Returns False indicate that given sequence does not exists."""
        return False

    def do_execute(
        self,
        cursor: Any,
        statement: str,
        parameters: Tuple[Any, ...],
        context: DefaultExecutionContext = None,
    ):
        cursor.execute(statement, parameters)

    def do_rollback(self, athena_connection: Connection):
        pass

    def set_isolation_level(self, athena_conn: Connection, level: str) -> None:
        pass

    def get_isolation_level(self, athena_conn: Connection) -> str:
        return "NONE"

    def get_default_isolation_level(self, athena_conn: Connection) -> str:
        return "NONE"  # TBD verify this is the correct value

    def _get_default_schema_name(self, connection: Connection) -> str:
        return ""

    def _handle_table_name(
        self, table_name: str, schema: Optional[str] = None
    ) -> Tuple[str, str]:
        return self._split_table_name(self._get_full_table(table_name, schema))

    def _split_table_name(self, full_table: str) -> Tuple[str, str]:
        parts = full_table.split(".")
        if len(parts) != 2:
            raise ValueError(f"Table name {full_table} is invalid.")
        return (parts[0], parts[1])

    def _get_full_table(
        self, table_name: str, schema: str = None, quote: bool = False
    ) -> str:
        res = (
            self.identifier_preparer.quote_identifier(table_name)
            if quote
            else table_name
        )
        if schema:
            schema_part = (
                self.identifier_preparer.quote_identifier(schema) if quote else schema
            )
            res = f"{schema_part}.{res}"

        if len(res.split(".")) == 3:
            raise ValueError(f"Table name {res} is invalid.")
        return res



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

import dash.development.base_component as bc
import dash_mantine_components as dmc
import mitzu.webapp.pages.explore.metric_type_handler as MTH

import mitzu.model as M
from dash import html

DATE_SELECTOR_DIV = "date_selector"
TIME_GROUP_DROPDOWN = "timegroup_dropdown"
LOOKBACK_WINDOW_DROPDOWN = "lookback_window_dropdown"
CUSTOM_DATE_PICKER = "custom_date_picker"
CUSTOM_DATE_TW_VALUE = "_custom_date"

CUSTOM_OPTION = {
    "label": "Custom",
    "value": CUSTOM_DATE_TW_VALUE,
}

DEFAULT_LOOKBACK_DAY_OPTION = M.TimeWindow(1, M.TimeGroup.MONTH)
LOOKBACK_DAYS_OPTIONS = {
    "Last 24 hours": M.TimeWindow(1, M.TimeGroup.DAY),
    "Last 7 days": M.TimeWindow(7, M.TimeGroup.DAY),
    "Last 14 days": M.TimeWindow(14, M.TimeGroup.DAY),
    "Last month": M.TimeWindow(1, M.TimeGroup.MONTH),
    "Last 2 month": M.TimeWindow(2, M.TimeGroup.MONTH),
    "Last 4 months": M.TimeWindow(4, M.TimeGroup.MONTH),
    "Last 6 months": M.TimeWindow(6, M.TimeGroup.MONTH),
    "Last 12 months": M.TimeWindow(12, M.TimeGroup.MONTH),
    "Last 2 years": M.TimeWindow(24, M.TimeGroup.MONTH),
    "Last 3 years": M.TimeWindow(36, M.TimeGroup.MONTH),
}


def get_time_group_options(exclude: List[M.TimeGroup]) -> List[Dict[str, Any]]:
    return [
        {"label": M.TimeGroup.group_by_string(tg), "value": tg.value}
        for tg in M.TimeGroup
        if tg not in exclude
    ]


def create_timewindow_options() -> List[Dict[str, str]]:
    return [{"label": k, "value": str(v)} for k, v in LOOKBACK_DAYS_OPTIONS.items()]


def get_default_tg_value(
    metric_config: M.MetricConfig, metric_type: MTH.MetricType
) -> M.TimeGroup:
    if metric_config is not None and metric_config.time_group is not None:
        return metric_config.time_group
    else:
        if metric_type == MTH.MetricType.RETENTION:
            return M.TimeGroup.WEEK
        else:
            return M.TimeGroup.DAY


def from_metric(
    metric_config: M.MetricConfig, metric_type: MTH.MetricType
) -> bc.Component:
    tg_val = get_default_tg_value(metric_config, metric_type)
    tw_options = create_timewindow_options()

    start_date = None
    end_date = None

    if metric_config is not None:
        if metric_config.start_dt is None and metric_config.lookback_days is not None:
            if metric_config.lookback_days in LOOKBACK_DAYS_OPTIONS.values():
                lookback_days = str(metric_config.lookback_days)
            else:
                lookback_days = str(DEFAULT_LOOKBACK_DAY_OPTION)
        elif metric_config.start_dt is not None:
            lookback_days = CUSTOM_DATE_TW_VALUE
        start_date = metric_config.start_dt
        end_date = metric_config.end_dt
    else:
        lookback_days = str(DEFAULT_LOOKBACK_DAY_OPTION)

    comp = html.Div(
        id=DATE_SELECTOR_DIV,
        children=[
            dmc.Select(
                id=TIME_GROUP_DROPDOWN,
                data=get_time_group_options(
                    exclude=[
                        M.TimeGroup.SECOND,
                        M.TimeGroup.MINUTE,
                        M.TimeGroup.QUARTER,
                    ]
                ),
                label="Period",
                value=tg_val.value,
                clearable=False,
                size="xs",
                style={"width": "120px"},
            ),
            html.Div(
                [
                    dmc.Select(
                        data=[CUSTOM_OPTION, *tw_options],
                        id=LOOKBACK_WINDOW_DROPDOWN,
                        className="me-1",
                        value=lookback_days,
                        clearable=False,
                        size="xs",
                        label="Time horizon",
                        searchable=True,
                        style={"width": "120px", "display": "inline-block"},
                    ),
                    dmc.DateRangePicker(
                        clearable=False,
                        inputFormat="YYYY-MM-DD",
                        id=CUSTOM_DATE_PICKER,
                        value=[start_date, end_date],
                        size="xs",
                        amountOfMonths=1,
                        style={
                            "width": "180px",
                            "display": "none"
                            if lookback_days != CUSTOM_DATE_TW_VALUE
                            else "inline-block",
                        },
                    ),
                ],
            ),
        ],
    )
    return comp


def get_metric_custom_dates(
    dd: Optional[M.DiscoveredProject], all_inputs: Dict[str, Any]
) -> Tuple[Optional[datetime], Optional[datetime]]:
    custom_dates = all_inputs.get(CUSTOM_DATE_PICKER, [None, None])
    start_dt = custom_dates[0]
    end_dt = custom_dates[1]
    lookback_days = all_inputs.get(LOOKBACK_WINDOW_DROPDOWN)

    if dd is None:
        return None, None

    if lookback_days == CUSTOM_DATE_TW_VALUE:
        if end_dt is None:
            end_dt = (
                dd.project.get_default_end_dt() if dd is not None else datetime.now()
            )
        if start_dt is None:
            def_lookback_window = (
                dd.project.webapp_settings.lookback_window.value
                if (
                    dd is not None
                    and dd.project.webapp_settings.lookback_window is not None
                )
                else 30
            )
            start_dt = end_dt - timedelta(days=def_lookback_window)
    else:
        return (None, None)
    return (start_dt, end_dt)


def get_metric_lookback_days(all_inputs: Dict[str, Any]) -> M.TimeWindow:
    tw_val = all_inputs.get(
        LOOKBACK_WINDOW_DROPDOWN,
    )
    if tw_val is None:
        return M.TimeWindow(1, M.TimeGroup.MONTH)
    if tw_val == CUSTOM_DATE_TW_VALUE:
        return M.DEF_LOOK_BACK_DAYS

    return M.TimeWindow.parse(str(tw_val))


def from_all_inputs(
    discovered_project: Optional[M.DiscoveredProject], all_inputs: Dict[str, Any]
) -> M.MetricConfig:
    lookback_days = get_metric_lookback_days(all_inputs)
    start_dt, end_dt = get_metric_custom_dates(discovered_project, all_inputs)
    time_group = M.TimeGroup(all_inputs.get(TIME_GROUP_DROPDOWN))

    return M.MetricConfig(
        start_dt=start_dt,
        end_dt=end_dt,
        lookback_days=lookback_days,
        time_group=time_group,
    )



from __future__ import annotations

from typing import Dict, List, Optional

import pandas as pd

import mitzu.model as M
import mitzu.webapp.cache as C
import mitzu.helper as H
import mitzu.webapp.model as WM
from mitzu.samples.data_ingestion import create_and_ingest_sample_project

SAMPLE_PROJECT_NAME = "sample_project"
SAMPLE_PROJECT_ID = "sample_project_id"
SAMPLE_CONNECTION_ID = "sample_connection_id"


DEFAULT_EXPIRE_TIME = 600  # 10 minutes

PROJECT_PREFIX = "__project__"
CONNECTION_PREFIX = "__conn__"
SEPC_PROJECT_PREFIX = "__spec_project__"
TABLE_DEFIONITION_PREFIX = "__table_definition__"
TABLE_PREFIX = "__table__"
SIMPLE_CHART_PREFIX = "__simple_chart__"
SAVED_METRIC_PREFIX = "__saved_metric__"
DASHBOARD_PREFIX = "__dashboard__"
EVENT_DEF_PREFIX = "__event_def__"


class InvalidStorageReference(Exception):
    pass


class StorageReference(M.Reference[M.ID]):
    def __init__(self, id: str, storage: MitzuStorage, prefix: str):
        super().__init__(None)
        self._id = id
        self.storage = storage
        self.prefix = prefix

    def get_id(self) -> Optional[str]:
        return super().get_id()

    def get_value(self) -> Optional[M.ID]:
        res = self._value_state.get_value()
        if self._id is None:
            return None

        if res is None:
            res = self.storage.mitzu_cache.get(self.prefix + self._id)
            H.LOGGER.debug(f"Restoring storage-ref value: {self.prefix + self._id}")
            self.restore_value(res)
        return res

    def get_value_if_exists(self) -> M.ID:
        return super().get_value_if_exists()

    def store_value(self):
        if self._id is not None:
            self.storage.mitzu_cache.clear(self.prefix + self._id)
        value = self._value_state.get_value()
        if value is not None:
            self.storage.mitzu_cache.put(self.prefix + value.get_id(), value)

    def set_value(self, value: Optional[M.ID]):
        super().set_value(value)
        self.store_value()

    def restore_value(self, value: Optional[M.ID]):
        super().restore_value(value)

    @classmethod
    def from_reference(
        cls,
        reference: M.Reference[M.ID],
        storage: MitzuStorage,
        prefix: str,
    ):
        if isinstance(reference, StorageReference):
            return reference

        ref_id = reference.get_id()
        if ref_id is None:
            return InvalidStorageReference(
                "Trying to convert Reference to StorageReference without an ID."
            )

        res: StorageReference[M.ID] = StorageReference(
            id=ref_id, storage=storage, prefix=prefix
        )
        res.restore_value(reference.get_value())
        return res


def setup_sample_project(storage: MitzuStorage):
    if storage.get_project(SAMPLE_PROJECT_ID) is not None:
        return
    connection = M.Connection(
        id=SAMPLE_CONNECTION_ID,
        connection_name="Sample connection",
        connection_type=M.ConnectionType.SQLITE,
        host="sample_project",
    )
    project = create_and_ingest_sample_project(
        connection,
        event_count=200000,
        number_of_users=300,
        schema="main",
        overwrite_records=False,
        project_id=SAMPLE_PROJECT_ID,
    )
    storage.set_connection(project.connection.id, project.connection)
    storage.set_project(project_id=project.id, project=project)

    dp = project.discover_project()

    for edt, defs in dp.definitions.items():
        storage.set_event_data_table_definition(
            project_id=project.id, definitions=defs, edt_full_name=edt.get_full_name()
        )


class MitzuStorage:
    def __init__(self, mitzu_cache: C.MitzuCache) -> None:
        super().__init__()
        self.mitzu_cache = mitzu_cache

    def set_project(self, project_id: str, project: M.Project):
        for edt in project.event_data_tables:
            edt.project_reference = StorageReference.from_reference(
                edt.project_reference, self, PROJECT_PREFIX
            )

        project._connection_ref = StorageReference.from_reference(
            project._connection_ref, self, CONNECTION_PREFIX
        )
        return self.mitzu_cache.put(PROJECT_PREFIX + project_id, project)

    def get_project(self, project_id: str) -> M.Project:
        project: M.Project = self.mitzu_cache.get(PROJECT_PREFIX + project_id)
        if project is not None:
            definitions: Dict[M.EventDataTable, Dict[str, M.Reference[M.EventDef]]] = {}
            for edt in project.event_data_tables:
                defs = self.get_event_data_table_definition(
                    project_id, edt.get_full_name()
                )
                if defs is not None:
                    definitions[edt] = defs

            project._discovered_project.set_value(
                M.DiscoveredProject(definitions, project)
            )
        return project

    def delete_project(self, project_id: str):
        self.mitzu_cache.clear(PROJECT_PREFIX + project_id)

    def list_projects(self) -> List[str]:
        return self.mitzu_cache.list_keys(PROJECT_PREFIX)

    def set_connection(self, connection_id: str, connection: M.Connection):
        self.mitzu_cache.put(CONNECTION_PREFIX + connection_id, connection)

    def get_connection(self, connection_id: str) -> M.Connection:
        return self.mitzu_cache.get(CONNECTION_PREFIX + connection_id)

    def delete_connection(self, connection_id: str):
        self.mitzu_cache.clear(CONNECTION_PREFIX + connection_id)

    def list_connections(self) -> List[str]:
        return self.mitzu_cache.list_keys(CONNECTION_PREFIX)

    def set_event_data_table_definition(
        self,
        project_id: str,
        edt_full_name: str,
        definitions: Dict[str, M.Reference[M.EventDef]],
    ):
        # Refernece can't be stored so we store the real values
        storage_refs: Dict[str, StorageReference[M.EventDef]] = {}
        for k, ref in definitions.items():
            st_ref = StorageReference.from_reference(
                reference=ref, storage=self, prefix=EVENT_DEF_PREFIX
            )
            storage_refs[k] = st_ref
            st_ref.store_value()

        self.mitzu_cache.put(
            SEPC_PROJECT_PREFIX + project_id + TABLE_DEFIONITION_PREFIX + edt_full_name,
            storage_refs,
        )

    def get_event_data_table_definition(
        self, project_id: str, edt_full_name: str
    ) -> Dict[str, M.Reference[M.EventDef]]:
        res = self.mitzu_cache.get(
            SEPC_PROJECT_PREFIX + project_id + TABLE_DEFIONITION_PREFIX + edt_full_name
        )

        return res

    def delete_event_data_table_definition(self, project_id: str, edt_full_name: str):
        self.mitzu_cache.clear(
            SEPC_PROJECT_PREFIX + project_id + TABLE_DEFIONITION_PREFIX + edt_full_name
        )

    def set_query_result_dataframe(
        self,
        metric_hash: str,
        dataframe: pd.DataFrame,
        expire: float = DEFAULT_EXPIRE_TIME,
    ):
        self.mitzu_cache.put(
            SIMPLE_CHART_PREFIX + metric_hash, dataframe, expire=expire
        )

    def get_query_result_dataframe(
        self,
        metric_hash: str,
    ) -> pd.DataFrame:
        return self.mitzu_cache.get(SIMPLE_CHART_PREFIX + metric_hash)

    def clear_query_result_dataframe(
        self,
        metric_hash: str,
    ):
        self.mitzu_cache.clear(SIMPLE_CHART_PREFIX + metric_hash)

    def set_saved_metric(self, metric_id: str, saved_metric: WM.SavedMetric):
        self.mitzu_cache.put(SAVED_METRIC_PREFIX + metric_id, saved_metric)

    def get_saved_metric(self, metric_id: str) -> WM.SavedMetric:
        res: WM.SavedMetric = self.mitzu_cache.get(SAVED_METRIC_PREFIX + metric_id)
        if res is not None:
            project = self.get_project(res.get_project_id())
            res.restore_project(project)
        return res

    def clear_saved_metric(self, metric_id: str):
        return self.mitzu_cache.clear(SAVED_METRIC_PREFIX + metric_id)

    def list_saved_metrics(self) -> List[str]:
        return self.mitzu_cache.list_keys(SAVED_METRIC_PREFIX)

    def list_dashboards(self) -> List[str]:
        return self.mitzu_cache.list_keys(DASHBOARD_PREFIX)

    def get_dashboard(self, dashboard_id: str) -> WM.Dashboard:
        dashboard: WM.Dashboard = self.mitzu_cache.get(DASHBOARD_PREFIX + dashboard_id)
        if dashboard is not None:
            for dm in dashboard.dashboard_metrics:
                sm = self.get_saved_metric(dm.get_saved_metric_id())
                dm.restore_saved_metric(sm)

        return dashboard

    def set_dashboard(self, dashboard_id: str, dashboard: WM.Dashboard):
        return self.mitzu_cache.put(DASHBOARD_PREFIX + dashboard_id, dashboard)

    def clear_dashboard(self, dashboard_id: str):
        return self.mitzu_cache.clear(DASHBOARD_PREFIX + dashboard_id)



from __future__ import annotations

import mitzu.adapters.generic_adapter as GA
import mitzu.model as M


def create_adapter(project: M.Project) -> GA.GenericDatasetAdapter:
    con_type = project.connection.connection_type
    res: GA.GenericDatasetAdapter
    if con_type == M.ConnectionType.FILE:
        from mitzu.adapters.file_adapter import FileAdapter

        res = FileAdapter(project)
    elif con_type == M.ConnectionType.SQLITE:
        from mitzu.adapters.file_adapter import SQLiteAdapter

        res = SQLiteAdapter(project)
    elif con_type == M.ConnectionType.ATHENA:
        from mitzu.adapters.athena_adapter import AthenaAdapter

        res = AthenaAdapter(project)
    elif con_type == M.ConnectionType.MYSQL:
        from mitzu.adapters.mysql_adapter import MySQLAdapter

        res = MySQLAdapter(project)
    elif con_type == M.ConnectionType.POSTGRESQL:
        from mitzu.adapters.postgresql_adapter import PostgresqlAdapter

        res = PostgresqlAdapter(project)
    elif con_type == M.ConnectionType.TRINO:
        from mitzu.adapters.trino_adapter import TrinoAdapter

        res = TrinoAdapter(project)
    elif con_type == M.ConnectionType.DATABRICKS:
        from mitzu.adapters.databricks_adapter import DatabricksAdapter

        res = DatabricksAdapter(project)
    elif con_type == M.ConnectionType.SNOWFLAKE:
        from mitzu.adapters.snowflake_adapter import SnowflakeAdapter

        res = SnowflakeAdapter(project)
    else:
        from mitzu.adapters.sqlalchemy_adapter import SQLAlchemyAdapter

        res = SQLAlchemyAdapter(project)

    return res



from __future__ import annotations

import re
import warnings
from typing import Any, Dict, List, cast

import mitzu.model as M

NUM_2_WORDS = {
    1: "one",
    2: "two",
    3: "three",
    4: "four",
    5: "five",
    6: "six",
    7: "seven",
    8: "eight",
    9: "nine",
}


def fix_def(val: str):
    fixed = re.sub("[^a-zA-Z0-9]", "_", val.lower())
    if fixed[0].isdigit():
        fixed = f"{NUM_2_WORDS[int(fixed[0])]}_{fixed[1:]}"
    return fixed[:32]


def _any_of(self: M.EventFieldDef, *vals: Any) -> M.SimpleSegment:
    return M.SimpleSegment(_left=self, _operator=M.Operator.ANY_OF, _right=vals)


def _not_any_of(self: M.EventFieldDef, *vals: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.NONE_OF,
        _right=vals,
    )


def _like(self: M.EventFieldDef, val: str) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.LIKE,
        _right=val,
    )


def _not_like(self: M.EventFieldDef, val: str) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.NOT_LIKE,
        _right=val,
    )


def _eq(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.EQ,
        _right=val,
    )


def _not_eq(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.NEQ,
        _right=val,
    )


def _gt(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.GT,
        _right=val,
    )


def _lt(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.LT,
        _right=val,
    )


def _gt_eq(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.GT_EQ,
        _right=val,
    )


def _lt_eq(self: M.EventFieldDef, val: Any) -> M.SimpleSegment:
    return M.SimpleSegment(
        _left=self,
        _operator=M.Operator.LT_EQ,
        _right=val,
    )


def add_enum_defs(
    event_field: M.EventFieldDef,
    class_def: Dict,
):
    if event_field._enums is not None:
        for enm in event_field._enums:
            class_def[f"{fix_def('is_'+str(enm))}"] = M.SimpleSegment(
                _left=event_field,
                _operator=M.Operator.EQ,
                _right=enm,
            )


class ModelLoader:
    def _create_event_field_class(
        self,
        event_name: str,
        event_field: M.EventFieldDef,
    ):
        event_field = event_field
        class_def: Dict[str, Any] = {}
        field_type = event_field._field._type

        if field_type == M.DataType.STRING:
            class_def["like"] = _like
            class_def["not_like"] = _not_like
            add_enum_defs(event_field, class_def)

        if field_type == M.DataType.BOOL:
            class_def["is_true"] = M.SimpleSegment(
                _left=event_field,
                _operator=M.Operator.EQ,
                _right=True,
            )
            class_def["is_false"] = M.SimpleSegment(
                _left=event_field,
                _operator=M.Operator.EQ,
                _right=False,
            )
        else:
            class_def["eq"] = _eq
            class_def["not_eq"] = _not_eq
            class_def["gt"] = _gt
            class_def["lt"] = _lt
            class_def["gt_eq"] = _gt_eq
            class_def["lt_eq"] = _lt_eq

            class_def["__eq__"] = _eq
            class_def["__ne__"] = _not_eq
            class_def["__gt__"] = _gt
            class_def["__lt__"] = _lt
            class_def["__ge__"] = _gt_eq
            class_def["__le__"] = _lt_eq
            class_def["not_any_of"] = _not_any_of
            class_def["any_of"] = _any_of

        class_def["is_null"] = M.SimpleSegment(
            _left=event_field, _operator=M.Operator.IS_NULL
        )
        class_def["is_not_null"] = M.SimpleSegment(
            _left=event_field, _operator=M.Operator.IS_NOT_NULL
        )

        return type(
            f"_{event_name}_{fix_def(event_field._field._name)}",
            (M.EventFieldDef,),
            class_def,
        )

    def _get_complex_ref(self, field: M.Field) -> List[str]:
        curr = field
        res: List[str] = []
        while curr._parent is not None:
            curr = curr._parent
            res.insert(0, curr._name)
        return res

    def _create_event_instance(self, event: M.EventDef):
        fields = event._fields

        class_def: Dict[str, Any] = {}
        for event_field, event_field_def in fields.items():

            field_class = self._create_event_field_class(
                event._event_name, event_field_def
            )
            class_instance = field_class(
                _event_name=event._event_name,
                _field=event_field,
                _event_data_table=event_field_def._event_data_table,
            )

            if event_field._parent is not None:
                field_name_chunks = self._get_complex_ref(event_field)
                curr_class = class_def
                type_name = f"_{event._event_name}"
                for fnc in field_name_chunks:
                    fnc = fix_def(fnc)
                    type_name = f"{type_name}_{fnc}"
                    if fnc not in curr_class:
                        curr_class[fnc] = type(type_name, (object,), {})()
                    curr_class = curr_class[fnc]

                class_field_name = fix_def(event_field._name)
                setattr(curr_class, class_field_name, class_instance)
            else:
                class_field_name = fix_def(event_field._name)
                class_def[class_field_name] = class_instance

        return type(f"_{event._event_name}", (M.SimpleSegment,), class_def)(event)

    def _process_schema(self, discovered_project: M.DiscoveredProject):
        project = discovered_project.project
        class_def = {}
        for ed_table in project.event_data_tables:
            definitions = discovered_project.definitions[ed_table]
            for event_name, event_def in definitions.items():
                fixed_name = fix_def(event_name)
                class_def[fixed_name] = self._create_event_instance(
                    event_def.get_value_if_exists()
                )

        return class_def

    def create_datasource_class_model(
        self, defs: M.DiscoveredProject
    ) -> M.DatasetModel:

        warnings.filterwarnings("ignore")
        class_def = self._process_schema(defs)
        return cast(
            M.DatasetModel, type("_dataset_model", (M.DatasetModel,), class_def)()
        )



from __future__ import annotations

from typing import Any, Dict, Optional, Tuple, cast
from urllib.parse import quote, urlparse, parse_qs

import dash_bootstrap_components as dbc
import dash.development.base_component as bc

import mitzu.model as M
import mitzu.webapp.model as WM
import mitzu.serialization as SE
import mitzu.helper as H

import mitzu.webapp.pages.explore.complex_segment_handler as CS
import mitzu.webapp.pages.explore.dates_selector_handler as DS
import mitzu.webapp.pages.explore.event_segment_handler as ES
import mitzu.webapp.pages.explore.graph_handler as GH
import mitzu.webapp.pages.explore.metric_config_handler as MC
import mitzu.webapp.pages.explore.metric_segments_handler as MS
import mitzu.webapp.pages.explore.metric_type_handler as MTH
import mitzu.webapp.pages.explore.simple_segment_handler as SS
import mitzu.webapp.pages.explore.toolbar_handler as TH
import mitzu.webapp.navbar as NB
import mitzu.visualization.charts as CHRT
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.storage as S
from mitzu.webapp.helper import MITZU_LOCATION, find_event_field_def, CHILDREN
import mitzu.webapp.pages.paths as P
import flask
import dash_mantine_components as dmc
import traceback
from dash import ctx, html, callback, no_update, dcc
from dash.dependencies import ALL, Input, Output, State
from dash_iconify import DashIconify
from mitzu.webapp.auth.decorator import restricted


from mitzu.webapp.helper import (
    get_final_all_inputs,
)

NAVBAR_ID = "explore_navbar"
SHARE_BUTTON = "share_button"
CLIPBOARD = "share_clipboard"
METRIC_NAME_INPUT = "metric_name_input"
METRIC_ID_VALUE = "metric_id_value"
METRIC_NAME_PROGRESS = "metric_name_progress"
EXPLORE_PAGE = "explore_page"
ALL_INPUT_COMPS = {
    "all_inputs": {
        ES.EVENT_NAME_DROPDOWN: Input(
            {"type": ES.EVENT_NAME_DROPDOWN, "index": ALL}, "value"
        ),
        SS.PROPERTY_OPERATOR_DROPDOWN: Input(
            {"type": SS.PROPERTY_OPERATOR_DROPDOWN, "index": ALL}, "value"
        ),
        SS.PROPERTY_NAME_DROPDOWN: Input(
            {"type": SS.PROPERTY_NAME_DROPDOWN, "index": ALL}, "value"
        ),
        SS.PROPERTY_VALUE_INPUT: Input(
            {"type": SS.PROPERTY_VALUE_INPUT, "index": ALL}, "value"
        ),
        CS.COMPLEX_SEGMENT_GROUP_BY: Input(
            {"type": CS.COMPLEX_SEGMENT_GROUP_BY, "index": ALL}, "value"
        ),
        DS.TIME_GROUP_DROPDOWN: Input(DS.TIME_GROUP_DROPDOWN, "value"),
        DS.CUSTOM_DATE_PICKER: Input(DS.CUSTOM_DATE_PICKER, "value"),
        DS.LOOKBACK_WINDOW_DROPDOWN: Input(DS.LOOKBACK_WINDOW_DROPDOWN, "value"),
        MC.TIME_WINDOW_INTERVAL_STEPS: Input(MC.TIME_WINDOW_INTERVAL_STEPS, "value"),
        MC.TIME_WINDOW_INTERVAL: Input(MC.TIME_WINDOW_INTERVAL, "value"),
        MC.AGGREGATION_TYPE: Input(MC.AGGREGATION_TYPE, "value"),
        MC.RESOLUTION_DD: Input(MC.RESOLUTION_DD, "value"),
        TH.GRAPH_REFRESH_BUTTON: Input(TH.GRAPH_REFRESH_BUTTON, "n_clicks"),
        TH.GRAPH_RUN_QUERY_BUTTON: Input(TH.GRAPH_RUN_QUERY_BUTTON, "n_clicks"),
        TH.CHART_TYPE_DD: Input(TH.CHART_TYPE_DD, "value"),
        TH.GRAPH_CONTENT_TYPE: Input(TH.GRAPH_CONTENT_TYPE, "value"),
        MTH.METRIC_TYPE_DROPDOWN: Input(MTH.METRIC_TYPE_DROPDOWN, "value"),
    }
}


def create_navbar(
    metric: Optional[M.Metric],
    saved_metric: Optional[WM.SavedMetric],
    project: M.Project,
    notebook_mode: bool,
) -> dbc.Navbar:
    navbar_children = [
        MTH.from_metric_type(MTH.MetricType.from_metric(metric)),
        dmc.TextInput(
            id=METRIC_NAME_INPUT,
            debounce=700,
            placeholder="Name your metric",
            value=saved_metric.name if saved_metric is not None else None,
            size="xs",
            icon=DashIconify(icon="material-symbols:star", color="dark"),
            rightSection=dmc.Loader(
                size="xs",
                color="dark",
                className="d-none",
                id=METRIC_NAME_PROGRESS,
            ),
            style={"width": "200px"},
        ),
        dbc.Button(
            [
                html.B(className="bi bi-link-45deg"),
                " Share",
                dcc.Clipboard(
                    id=CLIPBOARD,
                    content="",
                    className="position-absolute start-0 top-0 w-100 h-100 opacity-0",
                ),
            ],
            id=SHARE_BUTTON,
            className="position-relative top-0 start-0 text-nowrap",
            color="light",
            size="sm",
            style={"display": "none" if notebook_mode else "inline-block"},
        ),
    ]

    return NB.create_mitzu_navbar(
        id=NAVBAR_ID,
        children=navbar_children,
        off_canvas_toggler_visible=not notebook_mode,
        project_name=project.project_name,
    )


def create_explore_page(
    query_params: Dict[str, str],
    discovered_project: M.DiscoveredProject,
    storage: S.MitzuStorage,
    notebook_mode: bool = False,
) -> bc.Component:
    metric: Optional[M.Metric] = None
    saved_metric: Optional[WM.SavedMetric] = None
    if P.PROJECTS_EXPLORE_METRIC_QUERY in query_params:
        metric = SE.from_compressed_string(
            query_params[P.PROJECTS_EXPLORE_METRIC_QUERY], discovered_project.project
        )
    elif P.PROJECTS_EXPLORE_SAVED_METRIC_QUERY in query_params:
        saved_metric = storage.get_saved_metric(
            query_params[P.PROJECTS_EXPLORE_SAVED_METRIC_QUERY]
        )
        metric = saved_metric.metric

    metric_segments_div = MS.from_metric(metric, discovered_project)
    graph_container = create_graph_container(metric, discovered_project.project)
    navbar = create_navbar(
        metric=metric,
        saved_metric=saved_metric,
        project=discovered_project.project,
        notebook_mode=notebook_mode,
    )

    metric_id = H.create_unique_id() if saved_metric is None else saved_metric.id
    res = html.Div(
        children=[
            navbar,
            html.Div(children=metric_id, id=METRIC_ID_VALUE, className="d-none"),
            dbc.Container(
                children=[
                    dbc.Row(
                        children=[
                            dbc.Col(metric_segments_div, lg=4, md=12),
                            dbc.Col(graph_container, lg=8, md=12),
                        ],
                        justify="start",
                        align="top",
                        className="g-1",
                    ),
                ],
                fluid=True,
            ),
        ],
        className=EXPLORE_PAGE,
        id=EXPLORE_PAGE,
    )
    return res


def create_graph_container(metric: Optional[M.Metric], project: M.Project):
    metric_config = metric._config if metric is not None else M.MetricConfig()
    metric_type = MTH.MetricType.from_metric(metric)
    ret_or_conv_window = M.DEF_CONV_WINDOW
    if metric is not None:
        if isinstance(metric, M.ConversionMetric):
            ret_or_conv_window = metric._conv_window
        elif isinstance(metric, M.RetentionMetric):
            ret_or_conv_window = metric._retention_window
    else:
        if metric_type == MTH.MetricType.CONVERSION:
            ret_or_conv_window = M.DEF_CONV_WINDOW
        elif metric_type == MTH.MetricType.RETENTION:
            ret_or_conv_window = M.DEF_CONV_WINDOW

    metrics_config_card = MC.from_metric(metric_config, metric_type, ret_or_conv_window)
    graph_handler = GH.create_graph_container()
    toolbar_handler = TH.from_metric(metric, project)

    graph_container = dbc.Card(
        children=[
            dbc.CardBody(
                children=[
                    metrics_config_card,
                    toolbar_handler,
                    graph_handler,
                ],
            ),
        ],
    )
    return graph_container


def get_metric_group_by(
    metric_config: M.MetricConfig,
    metric_type: MTH.MetricType,
    all_inputs: Dict[str, Any],
    discovered_project: M.DiscoveredProject,
):
    group_by = None
    group_by_paths = all_inputs[MS.METRIC_SEGMENTS][CHILDREN]
    if len(group_by_paths) >= 1 and not (
        metric_type == MTH.MetricType.RETENTION
        and metric_config.time_group != M.TimeGroup.TOTAL
    ):
        gp = group_by_paths[0].get(CS.COMPLEX_SEGMENT_GROUP_BY)
        group_by = find_event_field_def(gp, discovered_project) if gp else None
        if group_by is not None:
            group_by._event_data_table.project_reference.restore_value(
                discovered_project.project
            )
    return group_by


def create_metric_from_all_inputs(
    all_inputs: Dict[str, Any],
    discovered_project: M.DiscoveredProject,
) -> Tuple[Optional[M.Metric], M.MetricConfig, M.TimeWindow]:
    segments = MS.from_all_inputs(discovered_project, all_inputs)
    metric_type = MTH.MetricType.parse(all_inputs[MTH.METRIC_TYPE_DROPDOWN])

    metric_config, time_window = MC.from_all_inputs(
        discovered_project, all_inputs, metric_type
    )
    metric: Optional[M.Metric] = None
    if metric_type == MTH.MetricType.CONVERSION:
        if len(segments) >= 1:
            metric = M.Conversion(segments)
    elif metric_type == MTH.MetricType.SEGMENTATION:
        if len(segments) == 1:
            metric = segments[0]
    elif metric_type == MTH.MetricType.RETENTION:
        if len(segments) == 2:
            metric = segments[0] >= segments[1]
        elif len(segments) == 1:
            metric = segments[0] >= segments[0]

    if metric is not None:
        metric = configure_metric(metric, metric_config, time_window)

    return metric, metric_config, time_window


def handle_input_changes(
    all_inputs: Dict[str, Any],
    discovered_project: M.DiscoveredProject,
) -> Dict[str, Any]:
    metric, metric_config, time_window = create_metric_from_all_inputs(
        all_inputs, discovered_project
    )
    parse_result = urlparse(all_inputs[MITZU_LOCATION])
    if metric is not None:

        url_params = "m=" + quote(SE.to_compressed_string(metric))
        parse_result = parse_result._replace(query=url_params)
    url = parse_result.geturl()

    metric_segments = MS.from_metric(
        discovered_project=discovered_project,
        metric=metric,
    ).children

    metric_type = all_inputs[MTH.METRIC_TYPE_DROPDOWN]

    mc_children = MC.from_metric(
        metric_config, MTH.MetricType.parse(metric_type), time_window
    ).children

    chart_type_dd = TH.create_chart_type_dropdown(metric)
    auto_refresh = discovered_project.project.webapp_settings.auto_refresh_enabled
    cancel_button_cls = (
        ""
        if (auto_refresh or ctx.triggered_id == TH.GRAPH_RUN_QUERY_BUTTON)
        else "d-none"
    )
    return {
        MS.METRIC_SEGMENTS: metric_segments,
        MC.METRICS_CONFIG_CONTAINER: mc_children,
        CLIPBOARD: url,
        MITZU_LOCATION: url,
        MTH.METRIC_TYPE_DROPDOWN: no_update,
        TH.CHART_TYPE_CONTAINER: chart_type_dd,
        TH.CANCEL_BUTTON: cancel_button_cls,
    }


def configure_metric(
    metric: M.Metric, metric_config: M.MetricConfig, time_window: M.TimeWindow
) -> M.Metric:
    if isinstance(metric, M.ConversionMetric):
        return M.ConversionMetric(
            conversion=metric._conversion,
            config=metric_config,
            conv_window=time_window,
        )

    if isinstance(metric, M.RetentionMetric):
        return M.RetentionMetric(
            initial_segment=metric._initial_segment,
            retaining_segment=metric._retaining_segment,
            retention_window=time_window,
            config=metric_config,
        )

    if isinstance(metric, M.SegmentationMetric):
        return M.SegmentationMetric(metric._segment, metric_config)

    raise Exception(f"Invalid metric type for graph visualization: {type(metric)}")


def create_callbacks():
    GH.create_callbacks()

    no_update_response = {
        MS.METRIC_SEGMENTS: no_update,
        MC.METRICS_CONFIG_CONTAINER: no_update,
        CLIPBOARD: no_update,
        MITZU_LOCATION: no_update,
        MTH.METRIC_TYPE_DROPDOWN: no_update,
        TH.CHART_TYPE_CONTAINER: no_update,
        TH.CANCEL_BUTTON: no_update,
    }

    @callback(
        output={
            MS.METRIC_SEGMENTS: Output(MS.METRIC_SEGMENTS, "children"),
            MC.METRICS_CONFIG_CONTAINER: Output(
                MC.METRICS_CONFIG_CONTAINER, "children"
            ),
            MITZU_LOCATION: Output(MITZU_LOCATION, "href"),
            CLIPBOARD: Output(CLIPBOARD, "content"),
            MTH.METRIC_TYPE_DROPDOWN: Output(MTH.METRIC_TYPE_DROPDOWN, "value"),
            TH.CHART_TYPE_CONTAINER: Output(TH.CHART_TYPE_CONTAINER, "children"),
            TH.CANCEL_BUTTON: Output(TH.CANCEL_BUTTON, "class_name"),
        },
        inputs=ALL_INPUT_COMPS,
        state=dict(
            href=State(MITZU_LOCATION, "href"),
            metric_name=State(METRIC_NAME_INPUT, "value"),
            metric_id=State(METRIC_ID_VALUE, "children"),
        ),
        prevent_initial_call=True,
    )
    @restricted
    def on_inputs_change(
        all_inputs: Dict[str, Any],
        href: str,
        metric_name: Optional[str],
        metric_id: str,
    ) -> Dict[str, Any]:
        try:
            if len(ctx.triggered_prop_ids) != 1:
                # Ignoring all components change as this comes from
                # page layout load and everything should be already rendered
                return no_update_response

            url = urlparse(href)
            project_id = P.get_path_value(
                P.PROJECTS_EXPLORE_PATH, url.path, P.PROJECT_ID_PATH_PART
            )
            depenedencies: DEPS.Dependencies = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            )
            project = depenedencies.storage.get_project(project_id)
            if project is None:
                return no_update_response
            all_inputs = get_final_all_inputs(all_inputs, ctx.inputs_list)
            all_inputs[METRIC_NAME_INPUT] = metric_name
            all_inputs[METRIC_ID_VALUE] = metric_id
            all_inputs[MITZU_LOCATION] = href
            discovered_project = project._discovered_project.get_value()
            if discovered_project is None:
                return no_update_response
            return handle_input_changes(all_inputs, discovered_project)
        except Exception:
            traceback.print_exc()
            return no_update_response


@callback(
    Output(METRIC_NAME_PROGRESS, "className"),
    Input(METRIC_NAME_INPUT, "value"),
    State(METRIC_ID_VALUE, "children"),
    State(MITZU_LOCATION, "href"),
    background=True,
    running=[
        (Output(METRIC_NAME_PROGRESS, "className"), "d-inline-block", "d-none"),
    ],
    prevent_initial_call=True,
)
@restricted
def handle_metric_name_changed(
    metric_name: str,
    metric_id: str,
    href: str,
) -> str:
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    parse_result = urlparse(href)
    sm = storage.get_saved_metric(metric_id)
    if sm is not None and sm.project is not None:
        storage.clear_saved_metric(metric_id)
        if metric_name:
            storage.set_saved_metric(metric_id, sm.rename(metric_name))
    else:

        project_id = P.get_path_value(
            P.PROJECTS_EXPLORE_PATH, parse_result.path, P.PROJECT_ID_PATH_PART
        )
        project = storage.get_project(project_id)
        metric_v64 = parse_qs(parse_result.query).get(P.PROJECTS_EXPLORE_METRIC_QUERY)
        if metric_v64 is not None:
            metric = SE.from_compressed_string(metric_v64[0], project)
            hash_key = GH.create_metric_hash_key(metric)
            result_df = GH.get_metric_result_df(hash_key, metric, storage)
            simple_chart = CHRT.get_simple_chart(metric, result_df)
            GH.store_rendered_saved_metric(
                metric_name=metric_name,
                metric=metric,
                simple_chart=simple_chart,
                project=project,
                storage=storage,
                metric_id=metric_id,
            )

    return "d-none"



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional
from uuid import uuid4

import mitzu.model as M
import mitzu.visualization.common as C
import mitzu.serialization as SE
from mitzu.helper import create_unique_id


@dataclass(frozen=True, init=False)
class SavedMetric(M.Identifiable):
    """
    SavedMetric class to store a group of a Metric and a SimpleChart

    :param chart: simple chart
    :param small_base64: image in base64 string format to represent the metric as thumbnail
    :param image_base64: larger image in base64 string format to represent the metric
    :param project: the project the metric can be queried on
    :param metric: metric
    :param saved_at: the time of creation
    """

    id: str
    name: str
    description: str
    chart: C.SimpleChart
    image_base64: str
    small_base64: str
    created_at: datetime
    last_updated_at: datetime
    owner: Optional[str]
    metric_json: str
    _project_ref: M.Reference[M.Project]
    _metric_state: M.State[
        M.Metric
    ]  # We shouldn't pickle the Metric as it won't have references to it's project.

    def __init__(
        self,
        name: str,
        chart: C.SimpleChart,
        image_base64: str,
        small_base64: str,
        project: M.Project,
        metric_json: Optional[str] = None,
        metric: Optional[M.Metric] = None,
        owner: Optional[str] = None,
        description: Optional[str] = None,
        created_at: Optional[datetime] = None,
        id: Optional[str] = None,
    ):
        if created_at is None:
            created_at = datetime.now()

        if metric_json is None:
            if metric is not None:
                metric_json = SE.to_compressed_string(metric)
            else:
                raise ValueError(
                    "Either metric or metric_json needs to be defined as an argument"
                )

        if id is None:
            id = create_unique_id()
        object.__setattr__(self, "chart", chart)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "description", description if description else "")
        object.__setattr__(self, "image_base64", image_base64)
        object.__setattr__(self, "small_base64", small_base64)
        object.__setattr__(self, "created_at", created_at)
        object.__setattr__(self, "last_updated_at", datetime.now())
        object.__setattr__(self, "metric_json", metric_json)
        object.__setattr__(self, "owner", owner)
        object.__setattr__(self, "id", id)
        object.__setattr__(self, "_project_ref", M.Reference(project))
        object.__setattr__(self, "_metric_state", M.State(metric))

    @property
    def project(self) -> Optional[M.Project]:
        return self._project_ref.get_value()

    @property
    def metric(self) -> Optional[M.Metric]:
        res = self._metric_state.get_value()
        if self.project is None:
            return None

        if res is None:
            res = SE.from_compressed_string(self.metric_json, self.project)
            self._metric_state.set_value(res)
        return res

    def restore_project(self, project: M.Project):
        self._project_ref.restore_value(project)

    def get_project_id(self) -> str:
        res = self._project_ref.get_id()
        if res is None:
            raise M.InvalidReferenceException("SavedMetric has no Project ID")
        return res

    def get_id(self) -> str:
        return self.id

    def rename(self, name: str) -> SavedMetric:
        if self.project is None:
            raise M.InvalidReferenceException(
                "Renaming saved metric however the project is missing"
            )
        return SavedMetric(
            name=name,
            id=self.id,
            metric_json=self.metric_json,
            description=self.description,
            owner=self.owner,
            chart=self.chart,
            image_base64=self.image_base64,
            small_base64=self.small_base64,
            created_at=self.created_at,
            project=self.project,
        )

    def refresh(
        self,
        image_base64: str,
        image_small64: str,
        metric: Optional[M.Metric] = None,
    ) -> SavedMetric:
        if self.project is None:
            raise M.InvalidReferenceException(
                "Refreshing saved metric however the project is missing"
            )
        return SavedMetric(
            name=self.name,
            id=self.id,
            metric=metric,
            project=self.project,
            description=self.description,
            owner=self.owner,
            chart=self.chart,
            image_base64=image_base64,
            small_base64=image_small64,
            created_at=self.created_at,
        )


@dataclass(init=False)
class DashboardMetric:
    """
    DashboardMetric class to store the positions of a Metric on the Dashboard

    :param saved_metric_id: the id of the corresponding saved_metric
    :param x: X pos
    :param y: Y pos
    :param width: Width
    :param height: Height
    :param saved_metric: The resolved saved_metric
    """

    x: int
    y: int
    width: int
    height: int
    _saved_metric_ref: M.Reference[SavedMetric]

    def __init__(
        self,
        saved_metric: SavedMetric,
        x: int = 0,
        y: int = 0,
        width: int = 2,
        height: int = 8,
    ):
        object.__setattr__(self, "x", x)
        object.__setattr__(self, "y", y)
        object.__setattr__(self, "width", width)
        object.__setattr__(self, "height", height)
        object.__setattr__(self, "_saved_metric_ref", M.Reference(saved_metric))

    @property
    def saved_metric(self) -> Optional[SavedMetric]:
        return self._saved_metric_ref.get_value()

    def restore_saved_metric(self, saved_metric: SavedMetric):
        self._saved_metric_ref.restore_value(saved_metric)

    def get_saved_metric_id(self) -> str:
        res = self._saved_metric_ref.get_id()
        if res is None:
            raise M.InvalidReferenceException("DashboardMetric has no SavedMetric ID")
        return res

    def set_saved_metric(self, saved_metric: SavedMetric):
        self._saved_metric_ref.set_value(saved_metric)


@dataclass(frozen=True)
class Dashboard:
    """
    Contains all details of a Dashboard.

    param name: the name of the dashboard
    param id: the id of the dashboard
    param dashboard_metric: list of dashboard metrics
    param created_at: the time of creation of the dashboard
    param owner: the name of the owner
    """

    name: str
    id: str = field(default_factory=lambda: str(uuid4())[-12:])
    dashboard_metrics: List[DashboardMetric] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.now)
    last_updated_at: Optional[datetime] = field(default_factory=datetime.now)
    owner: Optional[str] = None

    def update(self, dashboard_metrics: List[DashboardMetric]) -> Dashboard:
        return Dashboard(
            name=self.name,
            id=self.id,
            created_at=self.created_at,
            last_updated_at=datetime.now(),
            dashboard_metrics=dashboard_metrics,
            owner=self.owner,
        )

    def rename(self, name: str) -> Dashboard:
        return Dashboard(
            name=name,
            id=self.id,
            created_at=self.created_at,
            last_updated_at=self.last_updated_at,
            dashboard_metrics=self.dashboard_metrics,
            owner=self.owner,
        )



from __future__ import annotations

import flask
from typing import cast
from dash import (
    Input,
    Output,
    State,
    callback,
    html,
    register_page,
)
import dash_bootstrap_components as dbc
import dash.development.base_component as bc
import dash_mantine_components as dmc
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
import mitzu.webapp.service.user_service as US
from mitzu.webapp.helper import create_form_property_input
from mitzu.webapp.auth.decorator import (
    restricted_layout,
    restricted,
    restricted_for_admin,
)
from mitzu.webapp.webapp import MITZU_LOCATION

INDEX_TYPE = "user_property"
PROP_EMAIL = "email"
PROP_ROLE = "role"
PROP_PASSWORD = "password"
PROP_CONFIRM_PASSWORD = "confirm_password"


USER_SAVE_BUTTON = "user_save_button"
USER_CLOSE_BUTTON = "user_close_button"
NOT_FOUND_USER_CLOSE_BUTTON = "not_found_user_close_button"
USER_DELETE_BUTTON = "user_delete_button"
USER_CHANGE_PASSWORD_BUTTON = "user_change_password_button"
USER_CHANGE_ROLE_BUTTON = "user_change_role_button"

SAVE_RESPONSE_CONTAINER = "user_save_response_container"
DELETE_RESPONSE_CONTAINER = "user_delete_response_container"
CHANGE_PASSWORD_RESPONSE_CONTAINER = "user_change_password_response_container"
CHANGE_ROLE_RESPONSE_CONTAINER = "user_change_role_response_container"


@restricted_layout
def layout(user_id: str, **query_params) -> bc.Component:
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))
    user_service = deps.user_service
    if user_service is None:
        raise ValueError("User service is not set")

    if deps.authorizer is None:
        raise ValueError("Authorizer is not set")

    logged_in_user_id = deps.authorizer.get_current_user_id()
    if logged_in_user_id is None:
        raise ValueError("Cannot determine logged in user id")
    logged_in_user = user_service.get_user_by_id(logged_in_user_id)
    if logged_in_user is None:
        raise ValueError("Logged in user is not found")

    is_admin = deps.authorizer.get_current_user_role(flask.request) == US.Role.ADMIN
    show_password_fields = user_id == "new"
    show_change_password = False
    show_delete_button = user_id != "new"
    if user_id is not None and user_id != "new":
        if user_id == "my-account":
            user_id = logged_in_user.id

        show_change_password = (
            user_id == logged_in_user.id or logged_in_user.role == US.Role.ADMIN
        )
        user = user_service.get_user_by_id(user_id)
        show_delete_button = (
            logged_in_user.role == US.Role.ADMIN
            and user is not None
            and user.id != logged_in_user.id
        )
    else:
        user = None

    if user is None and user_id != "new":
        return html.Div(
            [
                NB.create_mitzu_navbar("users_edit", []),
                dbc.Container(
                    [
                        html.H4("Users not found"),
                        html.Hr(),
                        dbc.Button(
                            [html.B(className="bi bi-x"), " Close"],
                            color="secondary",
                            class_name="me-3",
                            id=NOT_FOUND_USER_CLOSE_BUTTON,
                            href=P.USERS_PATH,
                        ),
                    ]
                ),
            ]
        )

    return html.Div(
        [
            NB.create_mitzu_navbar("users_management_page", []),
            dbc.Container(
                [
                    html.H4("User"),
                    html.Hr(),
                    create_form_property_input(
                        index_type=INDEX_TYPE,
                        property=PROP_EMAIL,
                        icon_cls="bi bi-envelope",
                        type="text",
                        required=True,
                        value=user.email if user is not None else "",
                        read_only=user is not None,
                    ),
                    create_form_property_input(
                        index_type=INDEX_TYPE,
                        property=PROP_ROLE,
                        icon_cls="bi bi-person-fill",
                        component_type=dmc.Select,
                        data=[
                            {"label": v.name.lower(), "value": v.value}
                            for v in US.Role.all_values()
                        ],
                        required=True,
                        value=user.role if user is not None else US.Role.MEMBER.value,
                        read_only=user_service.is_root_user(user_id),
                    ),
                    create_form_property_input(
                        index_type=INDEX_TYPE,
                        property=PROP_PASSWORD,
                        icon_cls="bi bi-key",
                        type="password",
                        required=True,
                        value="",
                    )
                    if show_password_fields
                    else None,
                    create_form_property_input(
                        index_type=INDEX_TYPE,
                        property=PROP_CONFIRM_PASSWORD,
                        icon_cls="bi bi-key",
                        type="password",
                        required=True,
                        value="",
                    )
                    if show_password_fields
                    else None,
                ]
                + (
                    [
                        html.Hr(),
                        html.Div(
                            [
                                dbc.Button(
                                    [
                                        html.B(className="bi bi-check-circle me-1"),
                                        "Save",
                                    ],
                                    color="success",
                                    class_name="me-3",
                                    id=USER_SAVE_BUTTON,
                                ),
                                dbc.Button(
                                    [html.B(className="bi bi-x me-1"), "Delete"],
                                    color="danger",
                                    class_name="me-3",
                                    id=USER_DELETE_BUTTON,
                                    external_link=True,
                                    href=P.create_path(P.USERS_PATH),
                                )
                                if show_delete_button
                                else None,
                            ],
                            className="mb-3",
                        ),
                        html.Div(
                            children=[], id=SAVE_RESPONSE_CONTAINER, className="lead"
                        ),
                        html.Div(
                            children=[], id=DELETE_RESPONSE_CONTAINER, className="lead"
                        ),
                    ]
                    if user is None
                    else (
                        [
                            html.Hr(),
                            html.Div(
                                [
                                    dbc.Button(
                                        ["Update"],
                                        color="primary",
                                        class_name="me-3",
                                        id=USER_CHANGE_ROLE_BUTTON,
                                        disabled=user_service.is_root_user(user_id),
                                    )
                                    if user is not None
                                    else None,
                                ],
                                className="mb-3",
                            ),
                            html.Div(
                                children=[],
                                id=CHANGE_ROLE_RESPONSE_CONTAINER,
                                className="lead",
                            ),
                        ]
                        if is_admin
                        else []
                    )
                )
                + (change_password_form() if show_change_password else [])
                + [
                    html.Hr(),
                    dbc.Button(
                        [html.B(className="bi bi-x me-1"), "Close"],
                        color="secondary",
                        class_name="me-3",
                        id=USER_CLOSE_BUTTON,
                        href=P.USERS_PATH,
                    ),
                ],
            ),
        ]
    )


def change_password_form():
    return [
        html.Hr(),
        create_form_property_input(
            index_type=INDEX_TYPE,
            property=PROP_PASSWORD,
            icon_cls="bi bi-key",
            type="password",
            required=False,
            value="",
        ),
        create_form_property_input(
            index_type=INDEX_TYPE,
            property=PROP_CONFIRM_PASSWORD,
            icon_cls="bi bi-key",
            type="password",
            required=False,
            value="",
        ),
        html.Hr(),
        dbc.Button(
            ["Change password"],
            color="primary",
            className="mb-3",
            id=USER_CHANGE_PASSWORD_BUTTON,
        ),
        html.Div(children=[], id=CHANGE_PASSWORD_RESPONSE_CONTAINER, className="lead"),
    ]


@callback(
    output={
        SAVE_RESPONSE_CONTAINER: Output(SAVE_RESPONSE_CONTAINER, "children"),
    },
    inputs={
        "n_clicks": Input(USER_SAVE_BUTTON, "n_clicks"),
    },
    state={
        "email": State({"type": INDEX_TYPE, "index": PROP_EMAIL}, "value"),
        "role": State({"type": INDEX_TYPE, "index": PROP_ROLE}, "value"),
        "password": State({"type": INDEX_TYPE, "index": PROP_PASSWORD}, "value"),
        "confirm_password": State(
            {"type": INDEX_TYPE, "index": PROP_CONFIRM_PASSWORD}, "value"
        ),
    },
    prevent_initial_call=True,
)
@restricted_for_admin
def create_new_user(n_clicks: int, email="", role="", password="", confirm_password=""):
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))
    user_service = deps.user_service

    if user_service is None:
        raise ValueError("User service is not set")

    try:
        user_service.new_user(email, password, confirm_password, role=US.Role(role))
        return {
            SAVE_RESPONSE_CONTAINER: "User created!",
        }
    except Exception as e:
        return {
            SAVE_RESPONSE_CONTAINER: str(e),
        }


@callback(
    output={
        CHANGE_PASSWORD_RESPONSE_CONTAINER: Output(
            CHANGE_PASSWORD_RESPONSE_CONTAINER, "children"
        ),
    },
    inputs={
        "n_clicks": Input(USER_CHANGE_PASSWORD_BUTTON, "n_clicks"),
    },
    state={
        "password": State({"type": INDEX_TYPE, "index": PROP_PASSWORD}, "value"),
        "confirm_password": State(
            {"type": INDEX_TYPE, "index": PROP_CONFIRM_PASSWORD}, "value"
        ),
        "pathname": State(MITZU_LOCATION, "pathname"),
    },
    prevent_initial_call=True,
)
@restricted
def update_password(
    n_clicks: int, password="", confirm_password="", pathname: str = ""
):
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))

    user_service = deps.user_service

    if user_service is None:
        raise ValueError("User service is not set")

    if deps.authorizer is None:
        raise ValueError("Authorizer is not set")

    try:
        logged_in_user_id = deps.authorizer.get_current_user_id()
        logged_in_user = user_service.get_user_by_id(logged_in_user_id)

        user_id = P.get_path_value(P.USERS_HOME_PATH, pathname, P.USER_PATH_PART)
        if logged_in_user is None:
            raise Exception("User is not signed in")

        if user_id == "my-account":
            user_id = logged_in_user_id

        if logged_in_user is None:
            raise Exception("User is not signed in")

        if logged_in_user.role != US.Role.ADMIN and logged_in_user.id != user_id:
            raise Exception("User is not authorized to change this password")

        user_service.update_password(user_id, password, confirm_password)
        return {CHANGE_PASSWORD_RESPONSE_CONTAINER: "Password changed"}
    except Exception as e:
        return {
            CHANGE_PASSWORD_RESPONSE_CONTAINER: str(e),
        }


@callback(
    output={
        CHANGE_ROLE_RESPONSE_CONTAINER: Output(
            CHANGE_ROLE_RESPONSE_CONTAINER, "children"
        ),
    },
    inputs={
        "n_clicks": Input(USER_CHANGE_ROLE_BUTTON, "n_clicks"),
    },
    state={
        "role": State({"type": INDEX_TYPE, "index": PROP_ROLE}, "value"),
        "pathname": State(MITZU_LOCATION, "pathname"),
    },
    prevent_initial_call=True,
)
@restricted_for_admin
def update_role(n_clicks: int, role="", pathname: str = ""):
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))

    user_service = deps.user_service

    if user_service is None:
        raise ValueError("User service is not set")

    if deps.authorizer is None:
        raise ValueError("Authorizer is not set")

    try:
        logged_in_user_id = deps.authorizer.get_current_user_id()

        user_id = P.get_path_value(P.USERS_HOME_PATH, pathname, P.USER_PATH_PART)
        if logged_in_user_id is None:
            raise Exception("User is not signed in")

        if user_id == "my-account":
            user_id = logged_in_user_id

        user_service.update_role(user_id, US.Role(role))
        return {CHANGE_ROLE_RESPONSE_CONTAINER: "Role updated"}
    except Exception as e:
        return {
            CHANGE_ROLE_RESPONSE_CONTAINER: str(e),
        }


@callback(
    output={
        DELETE_RESPONSE_CONTAINER: Output(DELETE_RESPONSE_CONTAINER, "children"),
    },
    inputs={
        "n_clicks": Input(USER_DELETE_BUTTON, "n_clicks"),
    },
    state={
        "pathname": State(MITZU_LOCATION, "pathname"),
    },
    prevent_initial_call=True,
)
@restricted_for_admin
def delete_user(n_clicks: int, pathname: str = ""):
    user_id = P.get_path_value(P.USERS_HOME_PATH, pathname, P.USER_PATH_PART)
    deps = cast(DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY))
    user_service = deps.user_service

    if user_service is None:
        raise ValueError("User service is not set")

    try:
        if user_id == "my-account":
            raise Exception("Own account cannot be deleted")

        user_service.delete_user(user_id)
        return {
            DELETE_RESPONSE_CONTAINER: "User deleted",
        }

    except Exception as e:
        return {DELETE_RESPONSE_CONTAINER: str(e)}


register_page(
    __name__,
    path_template=P.USERS_HOME_PATH,
    title="Mitzu - Edit User",
    layout=layout,
)



from __future__ import annotations


from typing import Any, Dict

import dash_bootstrap_components as dbc
from dash import callback, ctx, html, dcc
from dash.dependencies import ALL, Input, Output
from mitzu import __version__ as version

import mitzu.webapp.configs as configs
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted

OFF_CANVAS = "off-canvas-id"
BUTTON_COLOR = "light"

CLOSE_BUTTON = "close-button"
SEARCH_INPUT = "search-input"

EXPLORE_BUTTON = "explore_button"
SAVED_METRICS_BUTTON = "saved_metrics"
DASHBOARDS_BUTTON = "dashboards_button"

EVENTS_PROPERTIES_BUTTON = "events_and_properties_button"
PROJECTS_BUTTON = "projects_button"
CONNECTIONS_BUTTON = "connections_button"

USERS_BUTTON = "users_button"
MY_ACCOUNT_BUTTON = "my_account_button"


MENU_ITEM_CSS = "mb-1 w-100 border-0 text-start"
EXPLORE_MENU_ITEM_CSS = "mb-1 w-100 text-start"


def create_offcanvas(dependencies: DEPS.Dependencies) -> dbc.Offcanvas:
    project_ids = dependencies.storage.list_projects()
    project_id = project_ids[0] if len(project_ids) > 0 else None

    show_users = False
    show_sign_out = False
    if dependencies.authorizer is not None:
        show_sign_out = True
        show_users = dependencies.authorizer._config.oauth is None

    res = dbc.Offcanvas(
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        dcc.Link(
                            html.Img(
                                src=configs.DASH_LOGO_PATH,
                                height="40px",
                                className="logo",
                            ),
                            href="/",
                        )
                    ),
                    dbc.Col(
                        dbc.Button(
                            html.I(className="bi bi-x-lg"),
                            color="dark",
                            id=CLOSE_BUTTON,
                            size="sm",
                            outline=True,
                            className="border-0 card-title pt-0",
                        ),
                        className="text-end",
                    ),
                ],
                className="mb-3",
            ),
            dbc.Input(
                id=SEARCH_INPUT,
                placeholder="Search ...",
                type="search",
                size="sm",
                className="mb-3 shadow-none",
            ),
            html.Hr(className="mb-3"),
            dbc.Button(
                [html.B(className="bi bi-search me-1"), "Explore"],
                color="info",
                class_name=MENU_ITEM_CSS,
                href=(
                    P.create_path(P.PROJECTS_EXPLORE_PATH, project_id=project_id)
                    if project_id is not None
                    else P.PROJECTS_PATH
                ),
                id=EXPLORE_BUTTON,
            ),
            dbc.Button(
                [html.B(className="bi bi-star-fill me-1"), "Saved metrics"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.SAVED_METRICS,
                id=SAVED_METRICS_BUTTON,
            ),
            dbc.Button(
                [html.B(className="bi bi-columns-gap me-1"), "Dashboards"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.DASHBOARDS_PATH,
                id=DASHBOARDS_BUTTON,
            ),
            html.Hr(className="mb-3"),
            dbc.Button(
                [html.B(className="bi bi-card-list me-1"), "Events and properties"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.EVENTS_AND_PROPERTIES_PATH,
                id=EVENTS_PROPERTIES_BUTTON,
            ),
            dbc.Button(
                [html.B(className="bi bi-play-circle me-1"), "Projects"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.PROJECTS_PATH,
                id=PROJECTS_BUTTON,
            ),
            dbc.Button(
                [html.B(className="bi bi-plugin me-1"), "Connections"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.CONNECTIONS_PATH,
                id=CONNECTIONS_BUTTON,
            ),
            html.Hr(className="mb-3"),
            dbc.Button(
                [html.B(className="bi bi-person-circle me-1"), "Users"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.USERS_PATH,
                id=USERS_BUTTON,
            )
            if show_users
            else None,
            dbc.Button(
                [html.B(className="bi bi-person-circle me-1"), "My Account"],
                color=BUTTON_COLOR,
                class_name=MENU_ITEM_CSS,
                href=P.create_path(P.USERS_HOME_PATH, user_id="my-account"),
                id=MY_ACCOUNT_BUTTON,
            )
            if show_users
            else None,
            html.Hr(className="mb-3 d-none"),
            dbc.Button(
                [html.B(className="bi bi-box-arrow-right me-1"), "Sign out"],
                color="light",
                class_name=MENU_ITEM_CSS,
                href=P.SIGN_OUT_URL,
                external_link=True,
            )
            if show_sign_out
            else None,
            html.Div(
                f"v{version}",
                className="position-absolute bottom-0 start-0 p-2",
                style={"opacity": "0.5"},
            ),
        ],
        close_button=False,
        is_open=False,
        id=OFF_CANVAS,
    )

    @callback(
        Output(OFF_CANVAS, "is_open"),
        inputs={
            "items": {
                NB.OFF_CANVAS_TOGGLER: Input(
                    {"type": NB.OFF_CANVAS_TOGGLER, "index": ALL}, "n_clicks"
                ),
                CLOSE_BUTTON: Input(CLOSE_BUTTON, "n_clicks"),
            }
        },
        prevent_initial_call=True,
    )
    @restricted
    def off_canvas_toggled(items: Dict[str, Any]) -> bool:
        for off_c in ctx.args_grouping["items"][NB.OFF_CANVAS_TOGGLER]:
            if off_c.get("triggered", False) and off_c.get("value") is not None:
                return True

        return False

    return res



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DatabricksQueryError(Exception):
    def __init__(self, error, query_id=None):
        self._error = error
        self._query_id = query_id

    @property
    def error_code(self):
        return self._error.get("errorCode", None)

    @property
    def error_name(self):
        return self._error.get("errorName", None)

    @property
    def error_type(self):
        return self._error.get("errorType", None)

    @property
    def error_exception(self):
        return self.failure_info.get("type", None) if self.failure_info else None

    @property
    def failure_info(self):
        return self._error.get("failureInfo", None)

    @property
    def message(self):
        return self._error.get("message", "Databricks did no return an error message")

    @property
    def error_location(self):
        location = self._error["errorLocation"]
        return (location["lineNumber"], location["columnNumber"])

    @property
    def query_id(self):
        return self._query_id

    def __repr__(self):
        return '{}(type={}, name={}, message="{}", query_id={})'.format(
            self.__class__.__name__,
            self.error_type,
            self.error_name,
            self.message,
            self.query_id,
        )

    def __str__(self):
        return repr(self)



from __future__ import annotations

from typing import Any, Dict, List, Optional

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import mitzu.model as M
import mitzu.webapp.pages.explore.event_segment_handler as ES
import mitzu.webapp.pages.explore.metric_type_handler as MTH
from dash import html
from mitzu.webapp.helper import (
    CHILDREN,
    get_event_names,
    get_property_name_label,
    WITH_VALUE_CLS,
)
import dash_mantine_components as dmc

COMPLEX_SEGMENT = "complex_segment"
COMPLEX_SEGMENT_BODY = "complex_segment_body"
COMPLEX_SEGMENT_FOOTER = "complex_segment_footer"
COMPLEX_SEGMENT_GROUP_BY = "complex_segment_group_by"


def get_group_by_options(
    discovered_project: M.DiscoveredProject, event_names: List[str]
):
    options: List[Dict[str, str]] = []
    for event_name in event_names:
        for field in discovered_project.get_event_def(event_name)._fields:
            field_value = field._get_name()
            should_break = False
            final_field_value = f"{event_name}.{field_value}"
            for op in options:
                if ".".join(op["value"].split(".")[1:]) == field_value:
                    should_break = True
                    break
            if not should_break:
                options.append(
                    {
                        "label": get_property_name_label(field._get_name()),
                        "value": final_field_value,
                    }
                )
    options.sort(key=lambda v: ".".join(v["value"].split(".")[1:]))
    return options


def create_group_by_dropdown(
    index: str,
    metric: Optional[M.Metric],
    segment: M.Segment,
    discovered_project: M.DiscoveredProject,
) -> dmc.Select:
    event_names = get_event_names(segment)
    group_by_efd = metric._config.group_by if metric is not None else None

    value = None
    if group_by_efd is not None:
        value = f"{group_by_efd._event_name}.{group_by_efd._field._get_name()}"

    options = get_group_by_options(discovered_project, event_names)
    if value not in [v["value"] for v in options]:
        value = None

    return dmc.Select(
        id={"type": COMPLEX_SEGMENT_GROUP_BY, "index": index},
        data=options,
        value=value,
        clearable=True,
        searchable=True,
        className=(
            COMPLEX_SEGMENT_GROUP_BY
            + " "
            + ("" if value is None else WITH_VALUE_CLS)
            + " border border-0 border-top-1 m-2 rounded-2"
        ),
        placeholder="+ Break Down",
    )


def find_all_event_segments(segment: M.Segment) -> List[M.Segment]:
    if segment is None:
        return []
    if isinstance(segment, M.SimpleSegment):
        return [segment]
    if isinstance(segment, M.ComplexSegment):
        if segment._operator == M.BinaryOperator.OR:
            return find_all_event_segments(segment._left) + find_all_event_segments(
                segment._right
            )
        else:
            return [segment]
    return []


def from_segment(
    discovered_project: M.DiscoveredProject,
    funnel_step: int,
    metric: Optional[M.Metric],
    segment: Optional[M.Segment],
    metric_type: MTH.MetricType,
) -> bc.Component:
    type_index = str(funnel_step)
    if metric_type == MTH.MetricType.CONVERSION:
        title = f"{funnel_step+1}. Step"
    elif metric_type == MTH.MetricType.RETENTION:
        title = "Initial Step" if funnel_step == 0 else "Retaining Step"
    else:
        title = "Events"

    header = dbc.CardHeader(title, class_name="p-2 fw-bold")

    body_children = []

    if segment is not None:
        event_segments = find_all_event_segments(segment)
        for index, evt_segment in enumerate(event_segments):
            body_children.append(
                ES.from_segment(evt_segment, discovered_project, funnel_step, index)
            )
    body_children.append(
        ES.from_segment(None, discovered_project, funnel_step, len(body_children))
    )

    if (
        segment is not None
        and funnel_step == 0
        and not (
            isinstance(metric, M.RetentionMetric)
            and metric._time_group != M.TimeGroup.TOTAL
        )
    ):
        group_by_dd = html.Div(
            [create_group_by_dropdown(type_index, metric, segment, discovered_project)],
            className=COMPLEX_SEGMENT_FOOTER,
        )
        body_children.append(group_by_dd)

    card_body = dbc.CardBody(
        children=body_children,
        className=COMPLEX_SEGMENT_BODY + " fw-normal p-0",
    )
    return dbc.Card(
        id={"type": COMPLEX_SEGMENT, "index": type_index},
        children=[header, card_body],
        className=COMPLEX_SEGMENT + " p-0",
    )


def from_all_inputs(
    discovered_project: M.DiscoveredProject,
    complex_segment: Dict[str, Any],
) -> Optional[M.Segment]:
    res_segment: Optional[M.Segment] = None
    event_segments = complex_segment.get(CHILDREN, {})

    for _, event_segment in event_segments.items():
        event_segment = ES.from_all_inputs(discovered_project, event_segment)
        if event_segment is None:
            continue
        if res_segment is None:
            res_segment = event_segment
        else:
            res_segment = res_segment | event_segment

    return res_segment



from __future__ import annotations

import mitzu.model as M
import mitzu.adapters.generic_adapter as GA
import pandas as pd
import mitzu.visualization.common as C


def fix_retention(pdf: pd.DataFrame, metric: M.RetentionMetric) -> pd.DataFrame:
    pdf[GA.AGG_VALUE_COL] = round(pdf[GA.AGG_VALUE_COL], 2)
    pdf[GA.GROUP_COL] = pdf[GA.GROUP_COL].fillna("n/a")
    if metric._time_group not in [
        M.TimeGroup.HOUR,
        M.TimeGroup.MINUTE,
        M.TimeGroup.SECOND,
        M.TimeGroup.TOTAL,
    ]:
        pdf[GA.DATETIME_COL] = pdf[GA.DATETIME_COL].dt.date
    return pdf


def get_retention_mapping(pdf: pd.DataFrame, metric: M.RetentionMetric) -> pd.DataFrame:
    if metric._time_group == M.TimeGroup.TOTAL:
        mapping = {
            GA.RETENTION_INDEX: C.X_AXIS_COL,
            GA.AGG_VALUE_COL: C.Y_AXIS_COL,
            GA.GROUP_COL: C.COLOR_COL,
        }
    else:
        if metric._group_by is not None:
            raise Exception(
                "Break downs are not supported for retention over time metric"
            )
        mapping = {
            GA.RETENTION_INDEX: C.X_AXIS_COL,
            GA.AGG_VALUE_COL: C.Y_AXIS_COL,
            GA.DATETIME_COL: C.COLOR_COL,
        }

    return pdf.rename(columns=mapping)



from __future__ import annotations

from typing import Dict, List, Optional, Callable

import mitzu.model as M
from mitzu.helper import LOGGER
from tqdm import tqdm
import sys
import traceback


class ProjectDiscoveryError(Exception):
    pass


class ProjectDiscovery:
    def __init__(self, project: M.Project, callback: Optional[Callable] = None):
        self.project = project
        self.callback = callback

    def _get_field_values(
        self,
        ed_table: M.EventDataTable,
        specific_fields: List[M.Field],
        event_specific: bool,
    ) -> Dict[str, M.EventDef]:
        return self.project.get_adapter().get_field_enums(
            event_data_table=ed_table,
            fields=specific_fields,
            event_specific=event_specific,
        )

    def _get_specific_fields(
        self, ed_table: M.EventDataTable, all_fields: List[M.Field]
    ) -> List[M.Field]:
        res = []
        if ed_table.event_name_field is None:
            return []
        if ed_table.event_specific_fields is None:
            res = [
                f
                for f in all_fields
                if (
                    f
                    not in (
                        ed_table.event_name_field,
                        ed_table.event_time_field,
                        ed_table.user_id_field,
                    )
                    and f._get_name() not in ed_table.ignored_fields
                )
            ]
        else:
            for spec_field in ed_table.event_specific_fields:
                res.extend(
                    [
                        f
                        for f in all_fields
                        if f._get_name().startswith(spec_field._get_name())
                    ]
                )
        return res

    def _copy_gen_field_def_to_spec(
        self, spec_event_name: str, gen_evt_field_def: M.EventFieldDef
    ):
        return M.EventFieldDef(
            _event_name=spec_event_name,
            _field=gen_evt_field_def._field,
            _event_data_table=gen_evt_field_def._event_data_table,
            _enums=gen_evt_field_def._enums,
        )

    def _merge_generic_and_specific_definitions(
        self,
        event_data_table: M.EventDataTable,
        generic: M.EventDef,
        specific: Dict[str, M.EventDef],
    ) -> Dict[str, M.Reference[M.EventDef]]:
        res: Dict[str, M.Reference[M.EventDef]] = {}
        for evt_name, spec_evt_def in specific.items():
            copied_gen_fields = {
                field: self._copy_gen_field_def_to_spec(evt_name, field_def)
                for field, field_def in generic._fields.items()
            }

            new_def = M.EventDef(
                _event_data_table=event_data_table,
                _event_name=evt_name,
                _fields={**spec_evt_def._fields, **copied_gen_fields},
            )
            res[evt_name] = M.Reference(new_def)

        return res

    def discover_project(self, progress_bar: bool = False) -> M.DiscoveredProject:
        definitions: Dict[M.EventDataTable, Dict[str, M.Reference[M.EventDef]]] = {}

        self.project.validate()
        tables = self.project.event_data_tables
        if progress_bar:
            tables = tqdm(
                tables, leave=False, file=sys.stdout, desc="Discovering datasets"
            )
        errors = {}
        for ed_table in tables:
            defs = {}
            try:
                LOGGER.debug(f"Discovering {ed_table.get_full_name()}")
                fields = self.project.get_adapter().list_fields(
                    event_data_table=ed_table
                )
                specific_fields = self._get_specific_fields(ed_table, fields)

                generic_fields = [c for c in fields if c not in specific_fields]
                generic_field_values = self._get_field_values(
                    ed_table, generic_fields, False
                )
                if M.ANY_EVENT_NAME not in generic_field_values.keys():
                    raise ProjectDiscoveryError(
                        f"Can't find any data to determine the generic field values in table '{ed_table.table_name}'"
                    )
                any_event_field_values = generic_field_values[M.ANY_EVENT_NAME]
                event_specific_field_values = self._get_field_values(
                    ed_table, specific_fields, True
                )
                defs = self._merge_generic_and_specific_definitions(
                    ed_table,
                    any_event_field_values,
                    event_specific_field_values,
                )
                definitions[ed_table] = defs
            except Exception as exc:
                LOGGER.error(f"{ed_table.table_name} failed to discover: {str(exc)}")
                errors[ed_table.table_name] = exc

            if self.callback is not None:
                self.callback(ed_table, defs, errors.get(ed_table.table_name))

        dd = M.DiscoveredProject(
            definitions=definitions,
            project=self.project,
        )

        if len(errors) > 0:
            traceback.print_exception(list(errors.values())[0])
            LOGGER.warning(f"Finished discovery with {len(errors)} errors.")
        else:
            LOGGER.info("Successfully finished dataset discovery.")
        return dd



from __future__ import annotations

import json
from datetime import datetime
from typing import cast, Optional

import mitzu.model as M
import pandas as pd
import sqlalchemy as SA
from mitzu.adapters.sqlalchemy_adapter import SQLAlchemyAdapter
from mitzu.helper import LOGGER, create_unique_id
from mitzu.samples.sample_data_generator import create_all_funnels
from tqdm import tqdm
import sys


def ingest_dataframe(
    engine: SA.engine.Engine,
    table_name: str,
    schema: str,
    df: pd.DataFrame,
    recreate: bool,
    show_progress: bool = False,
    chunk_size: int = 10000,
):
    ins = SA.inspect(engine)
    if ins.dialect.has_table(engine.connect(), table_name, schema=schema):
        LOGGER.warning(f"Table {table_name} already exists")
        if not recreate:
            return

    LOGGER.debug(f"Ingesting records into {table_name}")
    list_df = [
        df[i : i + chunk_size] for i in range(0, df.shape[0], chunk_size)  # noqa: E203
    ]

    mode = "replace"
    if show_progress:
        list_df = tqdm(
            list_df, leave=False, file=sys.stdout, desc=f"Populating {table_name}"
        )
    for df in list_df:
        df.to_sql(
            con=engine,
            name=table_name,
            schema=schema,
            index=False,
            if_exists=mode,
        )
        mode = "append"


def flatten_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    datetime_cols = []
    for col in df.dtypes.index:
        if df.dtypes[col].name == "datetime64[ns]":
            datetime_cols.append(col)

    json_struct = json.loads(df.to_json(orient="records"))
    norm_df = pd.json_normalize(json_struct)
    for col in datetime_cols:
        norm_df[col] = df[col]
    return norm_df


def create_and_ingest_sample_project(
    connection: M.Connection,
    schema: str = "main",
    event_count: int = 100000,
    number_of_users: int = 1000,
    overwrite_records: bool = True,
    show_progress: bool = False,
    chunk_size: int = 10000,
    seed: Optional[int] = 100,
    project_id: Optional[str] = None,
) -> M.Project:
    dfs = create_all_funnels(
        event_count=event_count, user_count=number_of_users, seed=seed
    )
    event_data_tables = []
    for table_name, df in dfs.items():
        event_data_tables.append(
            M.EventDataTable.create(
                table_name=table_name,
                schema=schema,
                event_time_field="event_time",
                user_id_field="user_id",
                event_name_alias=table_name if "event_name" not in df.columns else None,
                event_name_field="event_name" if "event_name" in df.columns else None,
            )
        )
    project = M.Project(
        connection=connection,
        project_name="Sample ecommerce project",
        event_data_tables=event_data_tables,
        discovery_settings=M.DiscoverySettings(
            lookback_days=365,
            end_dt=datetime(2022, 1, 1),
        ),
        webapp_settings=M.WebappSettings(
            end_date_config=M.WebappEndDateConfig.CUSTOM_DATE,
            custom_end_date=datetime(2022, 1, 1),
        ),
        project_id=project_id if project_id is not None else create_unique_id(),
    )

    for table_name, df in dfs.items():
        adapter = project.get_adapter()
        engine = cast(SQLAlchemyAdapter, adapter).get_engine()
        ingest_dataframe(
            engine,
            table_name=table_name,
            schema=schema,
            df=df,
            recreate=overwrite_records,
            show_progress=show_progress,
            chunk_size=chunk_size,
        )

    return project


if __name__ == "__main__":
    dfs = create_all_funnels(event_count=500000, user_count=2500)
    for name, df in dfs.items():
        df.to_parquet(f"{name}.parquet")



from __future__ import annotations

import mitzu.model as M

import mitzu.visualization.common as C
from typing import Any


def conv_time_suffix(max_ttc: int) -> str:
    return (
        "secs"
        if max_ttc <= C.TTC_RANGE_1_SEC
        else "mins"
        if max_ttc <= C.TTC_RANGE_2_SEC
        else "hours"
        if max_ttc <= C.TTC_RANGE_3_SEC
        else "days"
    )


def agg_type_label(agg_type: M.AggType, agg_param: Any = None):
    if agg_type == M.AggType.CONVERSION:
        return "Conversion rate"
    if agg_type == M.AggType.RETENTION_RATE:
        return "Retention rate"
    if agg_type == M.AggType.COUNT_UNIQUE_USERS:
        return "Unique users"
    if agg_type == M.AggType.COUNT_EVENTS:
        return "Event count"
    if agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
        return "Avg. time to convert"
    if agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
        return f"P{agg_param} time to convert"



from __future__ import annotations

import mitzu.model as M
import mitzu.samples.data_ingestion as DI
from typing import Optional


def get_sample_discovered_project(
    random_seed: Optional[int] = None,
) -> M.DiscoveredProject:
    """Generates a random discovered project, ready to be used.

    Args:
        random_seed (Optional[int], optional): the seed value for the random event generator. If None it defaults to a new seed every time.

    Returns:
        M.DiscoveredProject: The discovered project that is ready to be used.
    """
    connection = M.Connection(
        connection_type=M.ConnectionType.SQLITE, connection_name="Sample project"
    )

    project = DI.create_and_ingest_sample_project(connection, seed=random_seed)
    return project.discover_project()



import mitzu.adapters.generic_adapter as GA
import mitzu.model as M
from typing import Any, List, Dict
import pandas as pd
import mitzu.webapp.cache as CA
import hashlib

QUERY_PREFIX = "__QUERY__"
METRIC_PREFIX = "__METRIC__"


class CachingDatasetAdapter(GA.GenericDatasetAdapter):
    def __init__(
        self,
        adapter: GA.GenericDatasetAdapter,
        cache: CA.MitzuCache,
        expire: float = 600,
    ):
        self._adapter = adapter
        self._cache = cache
        self._expire = expire

    def execute_query(self, query: Any) -> pd.DataFrame:
        key = str(hashlib.md5(bytes(query)))
        pdf = self._cache.get(QUERY_PREFIX + key)
        if pdf is None:
            pdf = self._adapter.execute_query(query)
            self._cache.put(key, pdf, self._expire)
        return pdf

    def list_fields(self, event_data_table: M.EventDataTable) -> List[M.Field]:
        return self._adapter.list_fields(event_data_table)

    def get_distinct_event_names(self, event_data_table: M.EventDataTable) -> List[str]:

        return self._adapter.get_distinct_event_names(event_data_table)

    def get_field_enums(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> Dict[str, M.EventDef]:
        return self._adapter.get_field_enums(event_data_table, fields, event_specific)

    def get_conversion_sql(self, metric: M.ConversionMetric) -> str:
        return self._adapter.get_conversion_sql(metric)

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        key = hashlib.md5(metric.get_sql().encode()).hexdigest()
        pdf = self._cache.get(METRIC_PREFIX + key)
        if pdf is None:
            pdf = self._adapter.get_conversion_df(metric)
            self._cache.put(key, pdf, self._expire)
        return pdf

    def get_segmentation_sql(self, metric: M.SegmentationMetric) -> str:
        return self._adapter.get_segmentation_sql(metric)

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        key = hashlib.md5(metric.get_sql().encode()).hexdigest()
        pdf = self._cache.get(METRIC_PREFIX + key)
        if pdf is None:
            pdf = self._adapter.get_segmentation_df(metric)
            self._cache.put(key, pdf, self._expire)
        return pdf

    def get_retention_sql(self, metric: M.RetentionMetric) -> str:
        return self._adapter.get_retention_sql(metric)

    def get_retention_df(self, metric: M.RetentionMetric) -> pd.DataFrame:
        key = hashlib.md5(metric.get_sql().encode()).hexdigest()
        pdf = self._cache.get(METRIC_PREFIX + key)
        if pdf is None:
            pdf = self._adapter.get_retention_df(metric)
            self._cache.put(key, pdf, self._expire)
        return pdf

    def test_connection(self):
        self._adapter.test_connection()

    def stop_current_execution(self):
        self._adapter.stop_current_execution()
